from typing import List, Optional, Union

import numpy as np
import numpy.typing as npt
import pytest
import tensorflow as tf
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal
from tensorflow.keras import Input
from tensorflow.keras.layers import (
    Activation,
    Add,
    Average,
    Conv2D,
    Dense,
    Flatten,
    Input,
    Reshape,
)
from tensorflow.keras.models import Model, Sequential

from decomon.core import ForwardMode, Slope
from decomon.models.utils import ConvertMethod


@pytest.fixture(params=[m.value for m in ForwardMode])
def mode(request):
    return request.param


@pytest.fixture(params=[s.value for s in Slope])
def slope(request):
    return request.param


@pytest.fixture(params=list(range(10)))
def n(request):
    return request.param


@pytest.fixture(params=[32, 64, 16])
def floatx(request):
    # setup
    eps_bak = K.epsilon()  # store current epsilon
    floatx_bak = K.floatx()  # store current floatx
    precision = request.param
    K.set_floatx("float{}".format(precision))
    if precision == 16:
        K.set_epsilon(1e-2)
    # actual value
    yield precision
    # tear down
    K.set_epsilon(eps_bak)
    K.set_floatx(floatx_bak)


@pytest.fixture
def decimal(floatx):
    if floatx == 16:
        return 2
    else:
        return 4


@pytest.fixture(params=[True, False])
def dc_decomp(request):
    return request.param


@pytest.fixture(params=[True, False])
def use_bias(request):
    return request.param


@pytest.fixture(params=[True, False])
def shared(request):
    return request.param


@pytest.fixture(params=["same", "valid"])
def padding(request):
    return request.param


@pytest.fixture(params=[None, "linear", "relu"])
def activation(request):
    return request.param


@pytest.fixture(params=["channels_last", "channels_first"])
def data_format(request):
    return request.param


@pytest.fixture(params=[0, 1])
def odd(request):
    return request.param


@pytest.fixture(params=[1, 2, 3, -1])
def axis(request):
    return request.param


@pytest.fixture(params=[1, 2, 11])
def channels(request):
    return request.param


@pytest.fixture(params=[1, 2, 3])
def filter_size(request):
    return request.param


@pytest.fixture(params=[1, 2])
def strides(request):
    return request.param


@pytest.fixture(params=[True, False])
def flatten(request):
    return request.param


@pytest.fixture(params=[0, 1, 2, 3, 4, 5, 6])
def finetune_odd(request) -> Optional[np.ndarray]:
    # hard code several configuration of finetune for images odd=1 (6, 6, 2)
    finetune_params = np.zeros((6, 6, 2))
    if request.param == 1:
        finetune_params += 1
    elif request.param == 2:
        finetune_params[0] = 1
    elif request.param == 3:
        finetune_params[:, 0] = 1
    elif request.param == 4:
        finetune_params[:, :, 0] = 1
    elif request.param == 5:
        finetune_params[0, 0, 0] = 1
    else:
        return None

    return finetune_params


@pytest.fixture(params=[m.value for m in ConvertMethod])
def method(request):
    return request.param


class Helpers:
    @staticmethod
    def is_method_mode_compatible(method, mode):
        return not (
            ConvertMethod(method) in {ConvertMethod.CROWN_FORWARD_IBP, ConvertMethod.FORWARD_IBP}
            and ForwardMode(mode) != ForwardMode.IBP
        )

    @staticmethod
    def get_standard_values_1d_box(n, dc_decomp=True, grad_bounds=False, nb=100):
        """A set of functions with their monotonic decomposition for testing the activations"""
        w_u_ = np.ones(nb, dtype=K.floatx())
        b_u_ = np.zeros(nb, dtype=K.floatx())
        w_l_ = np.ones(nb, dtype=K.floatx())
        b_l_ = np.zeros(nb, dtype=K.floatx())

        if n == 0:
            # identity
            y_ = np.linspace(-2, -1, nb)
            x_ = np.linspace(-2, -1, nb)
            h_ = np.linspace(-2, -1, nb)
            g_ = np.zeros_like(x_)

        elif n == 1:
            y_ = np.linspace(1, 2, nb)
            x_ = np.linspace(1, 2, nb)
            h_ = np.linspace(1, 2, nb)
            g_ = np.zeros_like(x_)

        elif n == 2:
            y_ = np.linspace(-1, 1, nb)
            x_ = np.linspace(-1, 1, nb)
            h_ = np.linspace(-1, 1, nb)
            g_ = np.zeros_like(x_)

        elif n == 3:
            # identity
            y_ = np.linspace(-2, -1, nb)
            x_ = np.linspace(-2, -1, nb)
            h_ = 2 * np.linspace(-2, -1, nb)
            g_ = -np.linspace(-2, -1, nb)

        elif n == 4:
            y_ = np.linspace(1, 2, nb)
            x_ = np.linspace(1, 2, nb)
            h_ = 2 * np.linspace(1, 2, nb)
            g_ = -np.linspace(1, 2, nb)

        elif n == 5:
            y_ = np.linspace(-1, 1, nb)
            x_ = np.linspace(-1, 1, nb)
            h_ = 2 * np.linspace(-1, 1, nb)
            g_ = -np.linspace(-1, 1, nb)

        elif n == 6:
            assert nb == 100, "expected nb=100 samples"
            # cosine function
            x_ = np.linspace(-np.pi, np.pi, 100)
            y_ = np.cos(x_)
            h_ = np.concatenate([y_[:50], np.ones((50,))]) - 0.5
            g_ = np.concatenate([np.ones((50,)), y_[50:]]) - 0.5
            w_u_ = np.zeros_like(x_)
            w_l_ = np.zeros_like(x_)
            b_u_ = np.ones_like(x_)
            b_l_ = -np.ones_like(x_)

        elif n == 7:
            # h and g >0
            h_ = np.linspace(0.5, 2, nb)
            g_ = np.linspace(1, 2, nb)[::-1]
            x_ = h_ + g_
            y_ = h_ + g_

        elif n == 8:
            # h <0 and g <0
            # h_max+g_max <=0
            h_ = np.linspace(-2, -1, nb)
            g_ = np.linspace(-2, -1, nb)[::-1]
            y_ = h_ + g_
            x_ = h_ + g_

        elif n == 9:
            # h >0 and g <0
            # h_min+g_min >=0
            h_ = np.linspace(4, 5, nb)
            g_ = np.linspace(-2, -1, nb)[::-1]
            y_ = h_ + g_
            x_ = h_ + g_

        else:
            raise ValueError("n must be between 0 and 9.")

        x_min_ = x_.min() + np.zeros_like(x_)
        x_max_ = x_.max() + np.zeros_like(x_)

        x_0_ = np.concatenate([x_min_[:, None], x_max_[:, None]], 1)

        u_c_ = np.max(y_) * np.ones((nb,))
        l_c_ = np.min(y_) * np.ones((nb,))

        if dc_decomp:
            output = [
                x_[:, None],
                y_[:, None],
                x_0_[:, :, None],
                u_c_[:, None],
                w_u_[:, None, None],
                b_u_[:, None],
                l_c_[:, None],
                w_l_[:, None, None],
                b_l_[:, None],
                h_[:, None],
                g_[:, None],
            ]
        else:
            output = [
                x_[:, None],
                y_[:, None],
                x_0_[:, :, None],
                u_c_[:, None],
                w_u_[:, None, None],
                b_u_[:, None],
                l_c_[:, None],
                w_l_[:, None, None],
                b_l_[:, None],
            ]

        # cast element
        return [e.astype(K.floatx()) for e in output]

    @staticmethod
    def get_inputs_for_mode_from_full_inputs(
        inputs: Union[List[tf.Tensor], List[npt.NDArray[np.float_]]],
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        dc_decomp: bool = True,
    ) -> Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]:
        """Extract from full inputs the ones corresponding to the selected mode.

        Args:
            inputs: inputs from `get_standard_values_xxx()` or `get_tensor_decomposition_xxx()`
            mode:
            dc_decomp:

        Returns:

        """
        mode = ForwardMode(mode)
        if dc_decomp:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l, h, g = inputs
            if mode == ForwardMode.HYBRID:
                return inputs[2:]
            elif mode == ForwardMode.AFFINE:
                return [z, W_u, b_u, W_l, b_l, h, g]
            elif mode == ForwardMode.IBP:
                return [u_c, l_c, h, g]
            else:
                raise ValueError("Unknown mode.")
        else:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs
            if mode == ForwardMode.HYBRID:
                return inputs[2:]
            elif mode == ForwardMode.AFFINE:
                return [z, W_u, b_u, W_l, b_l]
            elif mode == ForwardMode.IBP:
                return [u_c, l_c]
            else:
                raise ValueError("Unknown mode.")

    @staticmethod
    def get_inputs_np_for_decomon_model_from_full_inputs(
        inputs: List[npt.NDArray[np.float_]],
    ) -> npt.NDArray[np.float_]:
        """Extract from full numpy inputs the ones for a decomon model prediction.

        Args:
            inputs:  inputs from `get_standard_values_xxx()`

        Returns:

        """
        l_c_, u_c_ = Helpers.get_input_ref_bounds_from_full_inputs(inputs=inputs)
        return np.concatenate((l_c_[:, None], u_c_[:, None]), axis=1)

    @staticmethod
    def get_input_ref_bounds_from_full_inputs(
        inputs: Union[List[tf.Tensor], List[npt.NDArray[np.float_]]],
    ) -> Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]:
        """Extract lower and upper bound for input ref from full inputs

        Args:
            inputs: inputs from `get_standard_values_xxx()` or `get_tensor_decomposition_xxx()`

        Returns:

        """
        u_c_, l_c_ = inputs[3], inputs[6]
        return [l_c_, u_c_]

    @staticmethod
    def prepare_full_np_inputs_for_convert_model(
        inputs: List[npt.NDArray[np.float_]],
        dc_decomp: bool = True,
    ) -> List[npt.NDArray[np.float_]]:
        """Prepare full numpy inputs for convert_forward or convert_backward.

        W_u and W_l will be idendity matrices, and b_u, b_l zeros vectors.

        Args:
            inputs: inputs from `get_standard_values_xxx()` or `get_tensor_decomposition_xxx()`

        Returns:

        """
        if dc_decomp:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l, h, g = inputs
        else:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs

        b_u = np.zeros_like(b_u)
        b_l = np.zeros_like(b_l)
        W_u = np.repeat(np.identity(n=W_u.shape[-1])[None, :, :], repeats=W_u.shape[0], axis=0)
        W_l = np.repeat(np.identity(n=W_l.shape[-1])[None, :, :], repeats=W_l.shape[0], axis=0)

        if dc_decomp:
            return [x, y, z, u_c, W_u, b_u, l_c, W_l, b_l, h, g]
        else:
            return [x, y, z, u_c, W_u, b_u, l_c, W_l, b_l]

    @staticmethod
    def get_input_tensors_for_decomon_convert_from_full_inputs(
        inputs: List[tf.Tensor],
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        dc_decomp: bool = True,
    ) -> List[tf.Tensor]:
        """Extract from full tensor inputs the ones for a conversion to decomon model.

        Args:
            inputs:  inputs from `get_tensor_decomposition_xxx()`
            mode:
            dc_decomp:

        Returns:

        """
        mode = ForwardMode(mode)
        if dc_decomp:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l, h, g = inputs
            input_box_tensor = K.concatenate([K.expand_dims(l_c, 1), K.expand_dims(u_c, 1)], 1)
            if mode == ForwardMode.HYBRID:
                return [input_box_tensor, u_c, W_u, b_u, l_c, W_l, b_l, h, g]
            elif mode == ForwardMode.AFFINE:
                return [input_box_tensor, W_u, b_u, W_l, b_l, h, g]
            elif mode == ForwardMode.IBP:
                return [u_c, l_c, h, g]
            else:
                raise ValueError("Unknown mode.")
        else:
            x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs
            input_box_tensor = K.concatenate([K.expand_dims(l_c, 1), K.expand_dims(u_c, 1)], 1)
            if mode == ForwardMode.HYBRID:
                return [input_box_tensor, u_c, W_u, b_u, l_c, W_l, b_l]
            elif mode == ForwardMode.AFFINE:
                return [input_box_tensor, W_u, b_u, W_l, b_l]
            elif mode == ForwardMode.IBP:
                return [u_c, l_c]
            else:
                raise ValueError("Unknown mode.")

    @staticmethod
    def get_input_ref_from_full_inputs(
        inputs: Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]
    ) -> Union[tf.Tensor, npt.NDArray[np.float_]]:
        """Extract from full inputs the input of reference for the original Keras layer.

        Args:
            inputs: inputs from `get_standard_values_xxx()` or `get_tensor_decomposition_xxx()`

        Returns:

        """
        return inputs[1]

    @staticmethod
    def get_tensor_decomposition_1d_box(dc_decomp=True):
        if dc_decomp:
            return [
                Input((1,), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((2, 1), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((1, 1), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((1, 1), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
                Input((1,), dtype=K.floatx()),
            ]
        return [
            Input((1,), dtype=K.floatx()),
            Input((1,), dtype=K.floatx()),
            Input((2, 1), dtype=K.floatx()),
            Input((1,), dtype=K.floatx()),
            Input((1, 1), dtype=K.floatx()),
            Input((1,), dtype=K.floatx()),
            Input((1,), dtype=K.floatx()),
            Input((1, 1), dtype=K.floatx()),
            Input((1,), dtype=K.floatx()),
        ]

    @staticmethod
    def get_full_outputs_from_outputs_for_mode(
        outputs_for_mode: Union[List[tf.Tensor], List[npt.NDArray[np.float_]]],
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        dc_decomp: bool = True,
        full_inputs: Optional[Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]] = None,
    ) -> Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]:
        mode = ForwardMode(mode)
        if dc_decomp:
            if mode == ForwardMode.HYBRID:
                z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = outputs_for_mode
            elif mode == ForwardMode.AFFINE:
                z_, w_u_, b_u_, w_l_, b_l_, h_, g_ = outputs_for_mode
                u_c_, l_c_ = None, None
            elif mode == ForwardMode.IBP:
                u_c_, l_c_, h_, g_ = outputs_for_mode
                z_, w_u_, b_u_, w_l_, b_l_ = None, None, None, None, None
                if full_inputs is not None:
                    z_ = full_inputs[2]
            else:
                raise ValueError("Unknown mode.")
        else:
            if mode == ForwardMode.HYBRID:
                z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_ = outputs_for_mode
            elif mode == ForwardMode.AFFINE:
                z_, w_u_, b_u_, w_l_, b_l_ = outputs_for_mode
                u_c_, l_c_ = None, None
            elif mode == ForwardMode.IBP:
                u_c_, l_c_ = outputs_for_mode
                z_, w_u_, b_u_, w_l_, b_l_ = None, None, None, None, None
                if full_inputs is not None:
                    z_ = full_inputs[2]
            else:
                raise ValueError("Unknown mode.")
            h_, g_ = None, None
        return [z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]

    @staticmethod
    def get_input_dim_multid_box(odd):
        if odd:
            return 3
        else:
            return 2

    @staticmethod
    def get_input_dim_images_box(odd):
        if odd:
            return 7
        else:
            return 6

    @staticmethod
    def get_input_dim_from_full_inputs(inputs: Union[List[tf.Tensor], List[npt.NDArray[np.float_]]]) -> int:
        """Get input_dim for to_decomon or to_backward from full inputs

        Args:
            inputs: inputs from `get_standard_values_xxx()` or `get_tensor_decomposition_xxx()`

        Returns:

        """
        return inputs[0].shape[-1]

    @staticmethod
    def get_tensor_decomposition_multid_box(odd=1, dc_decomp=True):
        n = Helpers.get_input_dim_multid_box(odd)

        if dc_decomp:
            # x, y, z, u, w_u, b_u, l, w_l, b_l, h, g
            return [
                Input((n,), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((2, n), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((n, n), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((n, n), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
                Input((n,), dtype=K.floatx()),
            ]
        return [
            Input((n,), dtype=K.floatx()),
            Input((n,), dtype=K.floatx()),
            Input((2, n), dtype=K.floatx()),
            Input((n,), dtype=K.floatx()),
            Input((n, n), dtype=K.floatx()),
            Input((n,), dtype=K.floatx()),
            Input((n,), dtype=K.floatx()),
            Input((n, n), dtype=K.floatx()),
            Input((n,), dtype=K.floatx()),
        ]

    @staticmethod
    def get_standard_values_multid_box(odd=1, dc_decomp=True):
        if dc_decomp:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
                h_0,
                g_0,
            ) = Helpers.get_standard_values_1d_box(0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
                h_1,
                g_1,
            ) = Helpers.get_standard_values_1d_box(1, dc_decomp)
        else:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
            ) = Helpers.get_standard_values_1d_box(0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
            ) = Helpers.get_standard_values_1d_box(1, dc_decomp)

        if not odd:
            # output (x_0+x_1, x_0+2*x_0)
            x_ = np.concatenate([x_0, x_1], -1)
            z_min_ = np.concatenate([z_0[:, 0], z_1[:, 0]], -1)
            z_max_ = np.concatenate([z_0[:, 1], z_1[:, 1]], -1)
            z_ = np.concatenate([z_min_[:, None], z_max_[:, None]], 1)
            y_ = np.concatenate([y_0 + y_1, y_0 + 2 * y_1], -1)
            b_u_ = np.concatenate([b_u_0 + b_u_1, b_u_0 + 2 * b_u_1], -1)
            u_c_ = np.concatenate([u_c_0 + u_c_1, u_c_0 + 2 * u_c_1], -1)
            b_l_ = np.concatenate([b_l_0 + b_l_1, b_l_0 + 2 * b_l_1], -1)
            l_c_ = np.concatenate([l_c_0 + l_c_1, l_c_0 + 2 * l_c_1], -1)

            if dc_decomp:
                h_ = np.concatenate([h_0 + h_1, h_0 + 2 * h_1], -1)
                g_ = np.concatenate([g_0 + g_1, g_0 + 2 * g_1], -1)

            w_u_ = np.zeros((len(x_), 2, 2))
            w_u_[:, 0, 0] = w_u_0[:, 0, 0]
            w_u_[:, 1, 0] = w_u_1[:, 0, 0]
            w_u_[:, 0, 1] = w_u_0[:, 0, 0]
            w_u_[:, 1, 1] = 2 * w_u_1[:, 0, 0]

            w_l_ = np.zeros((len(x_), 2, 2))
            w_l_[:, 0, 0] = w_l_0[:, 0, 0]
            w_l_[:, 1, 0] = w_l_1[:, 0, 0]
            w_l_[:, 0, 1] = w_l_0[:, 0, 0]
            w_l_[:, 1, 1] = 2 * w_l_1[:, 0, 0]

        else:
            (
                x_2,
                y_2,
                z_2,
                u_c_2,
                w_u_2,
                b_u_2,
                l_c_2,
                w_l_2,
                b_l_2,
                h_2,
                g_2,
            ) = Helpers.get_standard_values_1d_box(2)

            # output (x_0+x_1, x_0+2*x_0, x_2)
            x_ = np.concatenate([x_0, x_1, x_2], -1)
            z_min_ = np.concatenate([z_0[:, 0], z_1[:, 0], z_2[:, 0]], -1)
            z_max_ = np.concatenate([z_0[:, 1], z_1[:, 1], z_2[:, 1]], -1)
            z_ = np.concatenate([z_min_[:, None], z_max_[:, None]], 1)
            y_ = np.concatenate([y_0 + y_1, y_0 + 2 * y_1, y_2], -1)
            b_u_ = np.concatenate([b_u_0 + b_u_1, b_u_0 + 2 * b_u_1, b_u_2], -1)
            b_l_ = np.concatenate([b_l_0 + b_l_1, b_l_0 + 2 * b_l_1, b_l_2], -1)
            u_c_ = np.concatenate([u_c_0 + u_c_1, u_c_0 + 2 * u_c_1, u_c_2], -1)
            l_c_ = np.concatenate([l_c_0 + l_c_1, l_c_0 + 2 * l_c_1, l_c_2], -1)

            if dc_decomp:
                h_ = np.concatenate([h_0 + h_1, h_0 + 2 * h_1, h_2], -1)
                g_ = np.concatenate([g_0 + g_1, g_0 + 2 * g_1, g_2], -1)

            w_u_ = np.zeros((len(x_), 3, 3))
            w_u_[:, 0, 0] = w_u_0[:, 0, 0]
            w_u_[:, 1, 0] = w_u_1[:, 0, 0]
            w_u_[:, 0, 1] = w_u_0[:, 0, 0]
            w_u_[:, 1, 1] = 2 * w_u_1[:, 0, 0]
            w_u_[:, 2, 2] = w_u_2[:, 0, 0]

            w_l_ = np.zeros((len(x_), 3, 3))
            w_l_[:, 0, 0] = w_l_0[:, 0, 0]
            w_l_[:, 1, 0] = w_l_1[:, 0, 0]
            w_l_[:, 0, 1] = w_l_0[:, 0, 0]
            w_l_[:, 1, 1] = 2 * w_l_1[:, 0, 0]
            w_l_[:, 2, 2] = w_l_2[:, 0, 0]

        if dc_decomp:
            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]
        return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_]

    @staticmethod
    def build_image_from_1D_box(odd=0, m=0, dc_decomp=True):
        n = Helpers.get_input_dim_images_box(odd)

        if dc_decomp:
            (
                x_,
                y_0,
                z_,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
                h_0,
                g_0,
            ) = Helpers.get_standard_values_1d_box(m, dc_decomp=dc_decomp)
        else:
            (
                x_,
                y_0,
                z_,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
            ) = Helpers.get_standard_values_1d_box(m, dc_decomp=dc_decomp)

        y_ = np.concatenate([(i + 1) * y_0 for i in range(n * n)], -1).reshape((-1, n, n))
        b_u_ = np.concatenate([(i + 1) * b_u_0 for i in range(n * n)], -1).reshape((-1, n, n))
        b_l_ = np.concatenate([(i + 1) * b_l_0 for i in range(n * n)], -1).reshape((-1, n, n))

        if dc_decomp:
            h_ = np.concatenate([(i + 1) * h_0 for i in range(n * n)], -1).reshape((-1, n, n))
            g_ = np.concatenate([(i + 1) * g_0 for i in range(n * n)], -1).reshape((-1, n, n))

        u_c_ = np.concatenate([(i + 1) * u_c_0 for i in range(n * n)], -1).reshape((-1, n, n))
        l_c_ = np.concatenate([(i + 1) * l_c_0 for i in range(n * n)], -1).reshape((-1, n, n))

        w_u_ = np.zeros((len(x_), 1, n * n))
        w_l_ = np.zeros((len(x_), 1, n * n))

        for i in range(n * n):
            w_u_[:, 0, i] = (i + 1) * w_u_0[:, 0, 0]
            w_l_[:, 0, i] = (i + 1) * w_l_0[:, 0, 0]

        w_u_ = w_u_.reshape((-1, 1, n, n))
        w_l_ = w_l_.reshape((-1, 1, n, n))

        if dc_decomp:
            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]
        return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_]

    @staticmethod
    def build_image_from_2D_box(odd=0, m0=0, m1=1, dc_decomp=True):
        if dc_decomp:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
                h_0,
                g_0,
            ) = Helpers.build_image_from_1D_box(odd, m0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
                h_1,
                g_1,
            ) = Helpers.build_image_from_1D_box(odd, m1, dc_decomp)
        else:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
            ) = Helpers.build_image_from_1D_box(odd, m0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
            ) = Helpers.build_image_from_1D_box(odd, m1, dc_decomp)

        x_ = np.concatenate([x_0, x_1], -1)
        z_min_ = np.concatenate([z_0[:, 0], z_1[:, 0]], -1)
        z_max_ = np.concatenate([z_0[:, 1], z_1[:, 1]], -1)
        z_ = np.concatenate([z_min_[:, None], z_max_[:, None]], 1)
        y_ = y_0 + y_1
        b_u_ = b_u_0 + b_u_1
        b_l_ = b_l_0 + b_l_1

        u_c_ = u_c_0 + u_c_1
        l_c_ = l_c_0 + l_c_1

        w_u_ = np.concatenate([w_u_0, w_u_1], 1)
        w_l_ = np.concatenate([w_l_0, w_l_1], 1)

        if dc_decomp:
            h_ = h_0 + h_1
            g_ = g_0 + g_1

            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]
        return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_]

    @staticmethod
    def get_standard_values_images_box(data_format="channels_last", odd=0, m0=0, m1=1, dc_decomp=True):
        output = Helpers.build_image_from_2D_box(odd, m0, m1, dc_decomp)
        if dc_decomp:
            x_0, y_0, z_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0, h_0, g_0 = output
        else:
            x_0, y_0, z_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0 = output

        x_ = x_0
        z_ = z_0
        z_min_ = z_0[:, 0]
        z_max_ = z_0[:, 1]

        if data_format == "channels_last":
            y_0 = y_0[:, :, :, None]
            b_u_0 = b_u_0[:, :, :, None]
            b_l_0 = b_l_0[:, :, :, None]
            u_c_0 = u_c_0[:, :, :, None]
            l_c_0 = l_c_0[:, :, :, None]
            w_u_0 = w_u_0[:, :, :, :, None]
            w_l_0 = w_l_0[:, :, :, :, None]
            y_ = np.concatenate([y_0, y_0], -1)
            b_u_ = np.concatenate([b_u_0, b_u_0], -1)
            b_l_ = np.concatenate([b_l_0, b_l_0], -1)
            u_c_ = np.concatenate([u_c_0, u_c_0], -1)
            l_c_ = np.concatenate([l_c_0, l_c_0], -1)
            w_u_ = np.concatenate([w_u_0, w_u_0], -1)
            w_l_ = np.concatenate([w_l_0, w_l_0], -1)

            if dc_decomp:
                h_0 = h_0[:, :, :, None]
                g_0 = g_0[:, :, :, None]
                h_ = np.concatenate([h_0, h_0], -1)
                g_ = np.concatenate([g_0, g_0], -1)

        else:
            output = Helpers.get_standard_values_images_box(
                data_format="channels_last", odd=odd, m0=m0, m1=m1, dc_decomp=dc_decomp
            )

            x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_ = output[:9]
            if dc_decomp:
                h_, g_ = output[-2:]
                h_ = np.transpose(h_, (0, 3, 1, 2))
                g_ = np.transpose(g_, (0, 3, 1, 2))
            y_ = np.transpose(y_, (0, 3, 1, 2))
            u_c_ = np.transpose(u_c_, (0, 3, 1, 2))
            l_c_ = np.transpose(l_c_, (0, 3, 1, 2))
            b_u_ = np.transpose(b_u_, (0, 3, 1, 2))
            b_l_ = np.transpose(b_l_, (0, 3, 1, 2))
            w_u_ = np.transpose(w_u_, (0, 1, 4, 2, 3))
            w_l_ = np.transpose(w_l_, (0, 1, 4, 2, 3))

        if dc_decomp:
            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]
        else:
            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_]

    @staticmethod
    def get_tensor_decomposition_images_box(data_format, odd, dc_decomp=True):
        n = Helpers.get_input_dim_images_box(odd)

        if data_format == "channels_last":
            # x, y, z, u, w_u, b_u, l, w_l, b_l

            output = [
                Input((2,), dtype=K.floatx()),
                Input((n, n, 2), dtype=K.floatx()),
                Input((2, 2), dtype=K.floatx()),
                Input((n, n, 2), dtype=K.floatx()),
                Input((2, n, n, 2), dtype=K.floatx()),
                Input((n, n, 2), dtype=K.floatx()),
                Input((n, n, 2), dtype=K.floatx()),
                Input((2, n, n, 2), dtype=K.floatx()),
                Input((n, n, 2), dtype=K.floatx()),
            ]

            if dc_decomp:
                output += [Input((n, n, 2), dtype=K.floatx()), Input((n, n, 2), dtype=K.floatx())]
        else:
            output = [
                Input((2,), dtype=K.floatx()),
                Input((2, n, n), dtype=K.floatx()),
                Input((2, 2), dtype=K.floatx()),
                Input((2, n, n), dtype=K.floatx()),
                Input((2, 2, n, n), dtype=K.floatx()),
                Input((2, n, n), dtype=K.floatx()),
                Input((2, n, n), dtype=K.floatx()),
                Input((2, 2, n, n), dtype=K.floatx()),
                Input((2, n, n), dtype=K.floatx()),
            ]
            if dc_decomp:
                output += [Input((n, n, 2), dtype=K.floatx()), Input((n, n, 2), dtype=K.floatx())]

        return output

    @staticmethod
    def assert_output_properties_box(x_, y_, h_, g_, x_min_, x_max_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, decimal=4):
        if y_ is None:
            y_ = h_ + g_
        if h_ is not None:
            assert_almost_equal(h_ + g_, y_, decimal=decimal, err_msg="decomposition error")

        assert np.min(x_min_ <= x_max_), "x_min >x_max"

        assert_almost_equal(np.clip(x_min_ - x_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_min >x_")
        assert_almost_equal(np.clip(x_ - x_max_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_max < x_")
        if w_u_ is not None or w_l_ is not None:
            x_expand = x_ + np.zeros_like(x_)
            n_expand = len(w_u_.shape) - len(x_expand.shape)
            for i in range(n_expand):
                x_expand = np.expand_dims(x_expand, -1)

            if w_l_ is not None:
                lower_ = np.sum(w_l_ * x_expand, 1) + b_l_
            if w_u_ is not None:
                upper_ = np.sum(w_u_ * x_expand, 1) + b_u_

        # check that the functions h_ and g_ remains monotonic
        if h_ is not None:
            assert_almost_equal(
                np.clip(h_[:-1] - h_[1:], 0, np.inf),
                np.zeros_like(h_[1:]),
                decimal=decimal,
                err_msg="h is not increasing",
            )
            assert_almost_equal(
                np.clip(g_[1:] - g_[:-1], 0, np.inf),
                np.zeros_like(g_[1:]),
                decimal=decimal,
                err_msg="g is not increasing",
            )

        #
        if w_u_ is not None:
            if K.floatx() == "float32":
                assert_almost_equal(
                    np.clip(y_ - upper_, 0.0, 1e6),
                    np.zeros_like(y_),
                    decimal=decimal,
                    err_msg="upper <y",
                )
        if w_l_ is not None:
            if K.floatx() == "float32":
                assert_almost_equal(
                    np.clip(lower_ - y_, 0.0, np.inf),
                    np.zeros_like(y_),
                    decimal=decimal,
                    err_msg="lower_ >y",
                )

        if l_c_ is not None:
            assert_almost_equal(
                np.clip(l_c_ - y_, 0.0, np.inf),
                np.zeros_like(y_),
                decimal=decimal,
                err_msg="l_c >y",
            )
            assert_almost_equal(
                np.clip(y_ - u_c_, 0.0, 1e6),
                np.zeros_like(y_),
                decimal=decimal,
                err_msg="u_c <y",
            )

    @staticmethod
    def assert_decomon_model_output_properties_box(
        full_inputs, output_ref, outputs_for_mode, mode=ForwardMode.HYBRID, dc_decomp=True, decimal=4
    ):
        input_ref_ = Helpers.get_input_ref_from_full_inputs(full_inputs)
        input_ref_min_, input_ref_max_ = Helpers.get_input_ref_bounds_from_full_inputs(inputs=full_inputs)
        z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = Helpers.get_full_outputs_from_outputs_for_mode(
            outputs_for_mode=outputs_for_mode, mode=mode, dc_decomp=dc_decomp, full_inputs=full_inputs
        )

        Helpers.assert_output_properties_box(
            input_ref_,
            output_ref,
            h_,
            g_,
            input_ref_min_,
            input_ref_max_,
            u_c_,
            w_u_,
            b_u_,
            l_c_,
            w_l_,
            b_l_,
            decimal=decimal,
        )

    @staticmethod
    def assert_decomon_layer_output_properties_box(
        full_inputs, output_ref, outputs_for_mode, mode=ForwardMode.HYBRID, dc_decomp=True, decimal=4
    ):
        z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = Helpers.get_full_outputs_from_outputs_for_mode(
            outputs_for_mode=outputs_for_mode, mode=mode, dc_decomp=dc_decomp, full_inputs=full_inputs
        )
        x_ = full_inputs[0]
        Helpers.assert_output_properties_box(
            x_,
            output_ref,
            h_,
            g_,
            z_[:, 0],
            z_[:, 1],
            u_c_,
            w_u_,
            b_u_,
            l_c_,
            w_l_,
            b_l_,
            decimal=decimal,
        )

    @staticmethod
    def assert_decomon_layer_output_properties_box_linear(
        full_inputs, output_ref, outputs_for_mode, mode=ForwardMode.HYBRID, dc_decomp=True, decimal=4
    ):
        z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_ = Helpers.get_full_outputs_from_outputs_for_mode(
            outputs_for_mode=outputs_for_mode, mode=mode, dc_decomp=dc_decomp, full_inputs=full_inputs
        )
        x_ = full_inputs[0]
        Helpers.assert_output_properties_box_linear(
            x_,
            output_ref,
            z_[:, 0],
            z_[:, 1],
            u_c_,
            w_u_,
            b_u_,
            l_c_,
            w_l_,
            b_l_,
            decimal=decimal,
        )

    @staticmethod
    def assert_backward_layer_output_properties_box_linear(
        full_inputs, backward_outputs, output_ref=None, upper_constant_bound=None, lower_constant_bound=None, decimal=4
    ):
        w_u_, b_u_, w_l_, b_l_ = backward_outputs
        x_, y_, z_, u_c_, W_u_, B_u_, l_c_, W_l_, B_l_ = full_inputs

        # backward recomposition
        w_u_b = np.sum(np.maximum(w_u_, 0) * W_u_ + np.minimum(w_u_, 0) * W_l_, 1)[:, :, None]
        b_u_b = (
            b_u_ + np.sum(np.maximum(w_u_, 0) * B_u_[:, :, None], 1) + np.sum(np.minimum(w_u_, 0) * B_l_[:, :, None], 1)
        )
        w_l_b = np.sum(np.maximum(w_l_, 0) * W_l_ + np.minimum(w_l_, 0) * W_u_, 1)[:, :, None]
        b_l_b = (
            b_l_ + np.sum(np.maximum(w_l_, 0) * B_l_[:, :, None], 1) + np.sum(np.minimum(w_l_, 0) * B_u_[:, :, None], 1)
        )

        Helpers.assert_output_properties_box_linear(
            x_,
            output_ref,
            z_[:, 0],
            z_[:, 1],
            upper_constant_bound,
            w_u_b,
            b_u_b,
            lower_constant_bound,
            w_l_b,
            b_l_b,
            decimal=decimal,
        )

    @staticmethod
    def assert_output_properties_box_linear(x_, y_, x_min_, x_max_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, decimal=4):
        # flatten everyting
        n = len(x_)
        if y_ is not None:
            n = len(y_)
            y_ = y_.reshape((n, -1))
        if l_c_ is not None:
            u_c_ = u_c_.reshape((n, -1))
            l_c_ = l_c_.reshape((n, -1))
        if w_u_ is not None:
            w_u_ = w_u_.reshape((n, w_u_.shape[1], -1))
            w_l_ = w_l_.reshape((n, w_l_.shape[1], -1))
            b_u_ = b_u_.reshape((n, -1))
            b_l_ = b_l_.reshape((n, -1))

        assert np.min(x_min_ <= x_max_), "x_min >x_max"

        assert_almost_equal(np.clip(x_min_ - x_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_min >x_")
        assert_almost_equal(np.clip(x_ - x_max_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_max < x_")
        if w_u_ is not None:
            x_expand = x_ + np.zeros_like(x_)
            n_expand = len(w_u_.shape) - len(x_expand.shape)
            for i in range(n_expand):
                x_expand = np.expand_dims(x_expand, -1)

            lower_ = np.sum(w_l_ * x_expand, 1) + b_l_
            upper_ = np.sum(w_u_ * x_expand, 1) + b_u_

        if y_ is not None:
            if l_c_ is not None:
                assert_almost_equal(
                    np.clip(l_c_ - y_, 0.0, np.inf), np.zeros_like(y_), decimal=decimal, err_msg="l_c >y"
                )
                assert_almost_equal(np.clip(y_ - u_c_, 0.0, 1e6), np.zeros_like(y_), decimal=decimal, err_msg="u_c <y")
            if w_u_ is not None:
                assert_almost_equal(
                    np.clip(lower_ - y_, 0.0, np.inf), np.zeros_like(y_), decimal=decimal, err_msg="lower_ >y"
                )
                assert_almost_equal(
                    np.clip(y_ - upper_, 0.0, 1e6), np.zeros_like(y_), decimal=decimal, err_msg="upper <y"
                )

        # computer lower bounds on the domain
        if w_u_ is not None:
            x_expand_min = x_min_ + np.zeros_like(x_)
            x_expand_max = x_max_ + np.zeros_like(x_)
            n_expand = len(w_u_.shape) - len(x_expand_min.shape)
            for i in range(n_expand):
                x_expand_min = np.expand_dims(x_expand_min, -1)
                x_expand_max = np.expand_dims(x_expand_max, -1)

            lower_ = (
                np.sum(np.maximum(0, w_l_) * x_expand_min, 1) + np.sum(np.minimum(0, w_l_) * x_expand_max, 1) + b_l_
            )
            upper_ = (
                np.sum(np.maximum(0, w_u_) * x_expand_max, 1) + np.sum(np.minimum(0, w_u_) * x_expand_min, 1) + b_u_
            )

            if y_ is not None:
                assert_almost_equal(
                    np.clip(lower_ - y_, 0.0, np.inf), np.zeros_like(y_), decimal=decimal, err_msg="l_c >y"
                )
                assert_almost_equal(
                    np.clip(y_ - upper_, 0.0, 1e6), np.zeros_like(y_), decimal=decimal, err_msg="u_c <y"
                )

    # multi decomposition for convert
    @staticmethod
    def get_standard_values_multid_box_convert(odd=1, dc_decomp=True):
        if dc_decomp:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
                h_0,
                g_0,
            ) = Helpers.get_standard_values_1d_box(0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
                h_1,
                g_1,
            ) = Helpers.get_standard_values_1d_box(3, dc_decomp)
        else:
            (
                x_0,
                y_0,
                z_0,
                u_c_0,
                w_u_0,
                b_u_0,
                l_c_0,
                w_l_0,
                b_l_0,
            ) = Helpers.get_standard_values_1d_box(0, dc_decomp)
            (
                x_1,
                y_1,
                z_1,
                u_c_1,
                w_u_1,
                b_u_1,
                l_c_1,
                w_l_1,
                b_l_1,
            ) = Helpers.get_standard_values_1d_box(1, dc_decomp)

        if not odd:
            # output (x_0+x_1, x_0+2*x_0) (x_0, x_1)
            x_ = np.concatenate([x_0, x_1], -1)
            z_min_ = np.concatenate([z_0[:, 0], z_1[:, 0]], -1)
            z_max_ = np.concatenate([z_0[:, 1], z_1[:, 1]], -1)
            z_ = np.concatenate([z_min_[:, None], z_max_[:, None]], 1)
            y_ = np.concatenate([y_0, y_1], -1)
            b_u_ = np.concatenate([b_u_0, b_u_1], -1)
            u_c_ = np.concatenate([u_c_0, u_c_1], -1)
            b_l_ = np.concatenate([b_l_0, b_l_1], -1)
            l_c_ = np.concatenate([l_c_0, l_c_1], -1)

            if dc_decomp:
                h_ = np.concatenate([h_0, h_1], -1)
                g_ = np.concatenate([g_0, g_1], -1)

            w_u_ = np.zeros((len(x_), 2, 2))
            w_u_[:, 0, 0] = w_u_0[:, 0, 0]
            w_u_[:, 1, 1] = w_u_1[:, 0, 0]

            w_l_ = np.zeros((len(x_), 2, 2))
            w_l_[:, 0, 0] = w_l_0[:, 0, 0]
            w_l_[:, 1, 1] = w_l_1[:, 0, 0]

        else:
            if dc_decomp:
                (
                    x_2,
                    y_2,
                    z_2,
                    u_c_2,
                    w_u_2,
                    b_u_2,
                    l_c_2,
                    w_l_2,
                    b_l_2,
                    h_2,
                    g_2,
                ) = Helpers.get_standard_values_1d_box(2, dc_decomp)
            else:
                (x_2, y_2, z_2, u_c_2, w_u_2, b_u_2, l_c_2, w_l_2, b_l_2) = Helpers.get_standard_values_1d_box(
                    2, dc_decomp
                )
            x_ = np.concatenate([x_0, x_1, x_2], -1)
            z_min_ = np.concatenate([z_0[:, 0], z_1[:, 0], z_2[:, 0]], -1)
            z_max_ = np.concatenate([z_0[:, 1], z_1[:, 1], z_2[:, 1]], -1)
            z_ = np.concatenate([z_min_[:, None], z_max_[:, None]], 1)
            y_ = np.concatenate([y_0, y_1, y_2], -1)
            b_u_ = np.concatenate([b_u_0, b_u_1, b_u_2], -1)
            b_l_ = np.concatenate([b_l_0, b_l_1, b_l_2], -1)
            u_c_ = np.concatenate([u_c_0, u_c_1, u_c_2], -1)
            l_c_ = np.concatenate([l_c_0, l_c_1, l_c_2], -1)

            if dc_decomp:
                h_ = np.concatenate([h_0, h_1, h_2], -1)
                g_ = np.concatenate([g_0, g_1, g_2], -1)

            w_u_ = np.zeros((len(x_), 3, 3))
            w_u_[:, 0, 0] = w_u_0[:, 0, 0]
            w_u_[:, 1, 1] = w_u_1[:, 0, 0]
            w_u_[:, 2, 2] = w_u_2[:, 0, 0]

            w_l_ = np.zeros((len(x_), 3, 3))
            w_l_[:, 0, 0] = w_l_0[:, 0, 0]
            w_l_[:, 1, 1] = w_l_1[:, 0, 0]
            w_l_[:, 2, 2] = w_l_2[:, 0, 0]

        if dc_decomp:
            return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, h_, g_]
        return [x_, y_, z_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_]

    @staticmethod
    def assert_output_properties_box_nodc(x_, y_, x_min_, x_max_, u_c_, w_u_, b_u_, l_c_, w_l_, b_l_, decimal=4):
        assert np.min(x_min_ <= x_max_), "x_min >x_max"

        assert_almost_equal(np.clip(x_min_ - x_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_min >x_")
        assert_almost_equal(np.clip(x_ - x_max_, 0, np.inf), 0.0, decimal=decimal, err_msg="x_max < x_")

        x_expand = x_ + np.zeros_like(x_)
        n_expand = len(w_u_.shape) - len(x_expand.shape)
        for i in range(n_expand):
            x_expand = np.expand_dims(x_expand, -1)

        lower_ = np.sum(w_l_ * x_expand, 1) + b_l_
        upper_ = np.sum(w_u_ * x_expand, 1) + b_u_

        # check that the functions h_ and g_ remains monotonic

        assert_almost_equal(
            np.clip(l_c_ - y_, 0.0, np.inf),
            np.zeros_like(y_),
            decimal=decimal,
            err_msg="l_c >y",
        )
        assert_almost_equal(
            np.clip(y_ - u_c_, 0.0, 1e6),
            np.zeros_like(y_),
            decimal=decimal,
            err_msg="u_c <y",
        )

        #

        assert_almost_equal(
            np.clip(lower_ - y_, 0.0, np.inf),
            np.zeros_like(y_),
            decimal=decimal,
            err_msg="lower_ >y",
        )
        assert_almost_equal(
            np.clip(y_ - upper_, 0.0, 1e6),
            np.zeros_like(y_),
            decimal=decimal,
            err_msg="upper <y",
        )

        # computer lower bounds on the domain

        x_expand_min = x_min_ + np.zeros_like(x_)
        x_expand_max = x_max_ + np.zeros_like(x_)
        n_expand = len(w_u_.shape) - len(x_expand_min.shape)
        for i in range(n_expand):
            x_expand_min = np.expand_dims(x_expand_min, -1)
            x_expand_max = np.expand_dims(x_expand_max, -1)

        lower_ = np.sum(np.maximum(0, w_l_) * x_expand_min, 1) + np.sum(np.minimum(0, w_l_) * x_expand_max, 1) + b_l_
        upper_ = np.sum(np.maximum(0, w_u_) * x_expand_max, 1) + np.sum(np.minimum(0, w_u_) * x_expand_min, 1) + b_u_

        assert_almost_equal(
            np.clip(lower_.min(0) - l_c_.max(0), 0.0, np.inf),
            np.zeros_like(y_.min(0)),
            decimal=decimal,
            err_msg="lower_ >l_c",
        )
        assert_almost_equal(
            np.clip(u_c_.min(0) - upper_.max(0), 0.0, 1e6),
            np.zeros_like(y_.min(0)),
            decimal=decimal,
            err_msg="upper <u_c",
        )

    @staticmethod
    def toy_network_tutorial(dtype="float32"):
        layers = []
        layers.append(Dense(100, input_dim=1, dtype=dtype))  # specify the dimension of the input space
        layers.append(Activation("relu", dtype=dtype))
        layers.append(Dense(100, dtype=dtype))
        layers.append(Dense(1, activation="linear", dtype=dtype))
        model = Sequential(layers)
        return model

    @staticmethod
    def toy_network_tutorial_with_embedded_activation(dtype="float32"):
        layers = []
        layers.append(
            Dense(100, input_dim=1, activation="relu", dtype=dtype)
        )  # specify the dimension of the input space
        layers.append(Dense(100, dtype=dtype))
        layers.append(Dense(1, activation="linear", dtype=dtype))
        model = Sequential(layers)
        return model

    @staticmethod
    def toy_embedded_sequential(dtype="float32"):
        layers = []
        units = 10
        layers.append(
            Dense(units, input_dim=1, activation="relu", dtype=dtype)
        )  # specify the dimension of the input space
        layers.append(
            Helpers.dense_NN_1D(
                dtype=dtype, archi=[2, 3, 2], sequential=True, input_dim=units, activation="relu", use_bias=False
            )
        )
        layers.append(Dense(1, activation="linear", dtype=dtype))
        model = Sequential(layers)
        return model

    @staticmethod
    def dense_NN_1D(input_dim, archi, sequential, activation, use_bias, dtype="float32"):
        layers = [Dense(archi[0], use_bias=use_bias, activation=activation, input_dim=input_dim, dtype=dtype)]
        layers += [Dense(n_i, use_bias=use_bias, activation=activation, dtype=dtype) for n_i in archi[1:]]

        if sequential:
            return Sequential(layers)
        else:
            x = Input(input_dim, dtype=dtype)
            output = layers[0](x)
            for layer_ in layers[1:]:
                output = layer_(output)
            return Model(x, output)

    @staticmethod
    def toy_struct_v0_1D(input_dim, archi, activation, use_bias, merge_op=Add, dtype="float32"):
        nnet_0 = Helpers.dense_NN_1D(
            input_dim=input_dim, archi=archi, sequential=False, activation=activation, use_bias=use_bias, dtype=dtype
        )
        nnet_1 = Dense(archi[-1], use_bias=use_bias, activation="linear", input_dim=input_dim, name="toto", dtype=dtype)

        x = Input(input_dim, dtype=dtype)
        h_0 = nnet_0(x)
        h_1 = nnet_1(x)

        y = merge_op(dtype=dtype)([h_0, h_1])

        return Model(x, y)

    @staticmethod
    def toy_struct_v1_1D(input_dim, archi, sequential, activation, use_bias, merge_op=Add, dtype="float32"):
        nnet_0 = Helpers.dense_NN_1D(
            input_dim=input_dim,
            archi=archi,
            sequential=sequential,
            activation=activation,
            use_bias=use_bias,
            dtype=dtype,
        )

        x = Input(input_dim, dtype=dtype)
        h_0 = nnet_0(x)
        h_1 = nnet_0(x)
        y = merge_op(dtype=dtype)([h_0, h_1])

        return Model(x, y)

    @staticmethod
    def toy_struct_v2_1D(input_dim, archi, sequential, activation, use_bias, merge_op=Add, dtype="float32"):
        nnet_0 = Helpers.dense_NN_1D(
            input_dim=input_dim,
            archi=archi,
            sequential=sequential,
            activation=activation,
            use_bias=use_bias,
            dtype=dtype,
        )
        nnet_1 = Helpers.dense_NN_1D(
            input_dim=input_dim,
            archi=archi,
            sequential=sequential,
            activation=activation,
            use_bias=use_bias,
            dtype=dtype,
        )
        nnet_2 = Dense(archi[-1], use_bias=use_bias, activation="linear", input_dim=input_dim, dtype=dtype)

        x = Input(input_dim, dtype=dtype)
        nnet_0(x)
        nnet_1(x)
        nnet_1.set_weights([-p for p in nnet_0.get_weights()])  # be sure that the produced output will differ
        h_0 = nnet_2(nnet_0(x))
        h_1 = nnet_2(nnet_1(x))
        y = merge_op(dtype=dtype)([h_0, h_1])

        return Model(x, y)

    @staticmethod
    def toy_struct_cnn(dtype="float32", image_data_shape=(6, 6, 2)):
        input_dim = np.prod(image_data_shape)
        layers = [
            Reshape(target_shape=image_data_shape, input_dim=input_dim),
            Conv2D(
                10,
                kernel_size=(3, 3),
                activation="relu",
                data_format="channels_last",
                dtype=dtype,
            ),
            Flatten(dtype=dtype),
            Dense(1, dtype=dtype),
        ]
        return Sequential(layers)

    @staticmethod
    def toy_model(model_name, dtype="float32"):
        if model_name == "tutorial":
            return Helpers.toy_network_tutorial(dtype=dtype)
        elif model_name == "tutorial_activation_embedded":
            return Helpers.toy_network_tutorial_with_embedded_activation(dtype=dtype)
        elif model_name == "merge_v0":
            return Helpers.toy_struct_v0_1D(dtype=dtype, input_dim=1, archi=[2, 3, 2], activation="relu", use_bias=True)
        elif model_name == "merge_v1":
            return Helpers.toy_struct_v1_1D(
                dtype=dtype, input_dim=1, archi=[2, 3, 2], activation="relu", use_bias=True, sequential=False
            )
        elif model_name == "merge_v1_seq":
            return Helpers.toy_struct_v1_1D(
                dtype=dtype, input_dim=1, archi=[2, 3, 2], activation="relu", use_bias=True, sequential=True
            )
        elif model_name == "merge_v1_2":
            return Helpers.toy_struct_v1_1D(
                dtype=dtype, input_dim=2, archi=[2, 3, 2], activation="relu", use_bias=True, sequential=False
            )
        elif model_name == "merge_v2":
            return Helpers.toy_struct_v2_1D(
                dtype=dtype, input_dim=1, archi=[2, 3, 2], activation="relu", use_bias=True, sequential=False
            )
        elif model_name == "cnn":
            return Helpers.toy_struct_cnn(dtype=dtype)
        elif model_name == "embedded_model":
            return Helpers.toy_embedded_sequential(dtype=dtype)
        else:
            raise ValueError(f"model_name {model_name} unknown")


@pytest.fixture
def helpers():
    return Helpers


@pytest.fixture(
    params=[
        "tutorial",
        "tutorial_activation_embedded",
        "merge_v0",
        "merge_v1",
        "merge_v1_seq",
        "merge_v1_2",
        "merge_v2",
        "cnn",
        "embedded_model",
    ]
)
def toy_model(request, helpers):
    model_name = request.param
    return helpers.toy_model(model_name, dtype=K.floatx())


@pytest.fixture(
    params=[
        "tutorial",
        "tutorial_activation_embedded",
        "merge_v0",
        "merge_v1",
        "merge_v1_seq",
        "merge_v2",
        "embedded_model",
    ]
)
def toy_model_1d(request, helpers):
    model_name = request.param
    return helpers.toy_model(model_name, dtype=K.floatx())



from typing import Any, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K

from decomon.core import (
    BoxDomain,
    ForwardMode,
    InputsOutputsSpec,
    PerturbationDomain,
    get_affine,
)

# step 1: compute (x_i, y_i) such that x_i[j]=l_j if j==i else u_j
# dataset of size n+1 on which we can compute an affine bound


def get_upper_linear_hull_max(
    inputs: List[tf.Tensor],
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    perturbation_domain: Optional[PerturbationDomain] = None,
    axis: int = -1,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Compute the linear hull that overapproximates max along the axis dimension

    Args:
        inputs: list of input tensors
        mode: type of Forward propagation (ibp, affine, or hybrid). Default to hybrid.
        perturbation_domain (optional): type of perturbation domain that encompass the set of perturbations. Defaults to None.
        axis (optional): Defaults to -1. See Keras offical documentation backend.max(., axis)

    Raises:
        NotImplementedError: axis <0 and axis!=-1

    Returns:
        list of output tensors. The upper linear relaxation of max(., axis) in the mode format
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    input_shape = inputs_outputs_spec.get_input_shape(inputs)

    # attention if axis=-1 or axis=n
    dtype32 = "float32"
    o_value = K.cast(1.0, dtype)
    z_value = K.cast(0.0, dtype)
    if axis == len(input_shape):
        axis = -1
    if axis != -1 and axis < 0:
        raise NotImplementedError()  # to do

    # get the shape of the dimension
    n_dim = input_shape[axis]

    # expand dim/broadcast
    mask = tf.linalg.diag(tf.ones((n_dim), dtype=dtype))  # (n_dim, n_dim)
    mask_shape = np.ones(len(u_c.shape) + 1)
    mask_shape[-1] = n_dim
    if axis != -1:
        mask_shape[axis] = n_dim
    else:
        mask_shape[-2] = n_dim
    mask = K.reshape(mask, mask_shape)  # (1, shape, n_dim, n_dim)

    l_c_reshaped = K.expand_dims(l_c, -1)  # (1, shape, n_dim, 1)
    u_c_reshaped = K.expand_dims(u_c, -1)  # (1, shape, n_dim, 1)

    # detect collapsed dimensions: lower[i]==upper[i]
    index_collapse = o_value - K.sign(l_c_reshaped - u_c_reshaped)

    corners = mask * l_c_reshaped + (o_value - mask) * (u_c_reshaped)  # (1, shape, n_dim, n_dim)
    corners_collapse = mask * l_c_reshaped + (o_value - mask) * (u_c_reshaped + index_collapse)
    # add the corners containing all the upper bounds
    corners_collapse = K.concatenate(
        [corners_collapse, u_c_reshaped + index_collapse], axis=-1
    )  # (None, shape, n_dim, n_dim+1)
    corners = K.concatenate([corners, u_c_reshaped], axis=-1)

    if axis != -1:
        corners_pred = K.max(corners, axis=axis)  # (None, shape_, n_dim+1)
    else:
        corners_pred = K.max(corners, axis=-2)  # (None, shape_, n_dim+1)

    # include bias in corners
    if axis != -1:
        bias_corner = o_value + tf.math.reduce_sum(z_value * corners, axis, keepdims=True)
        corners_collapse = K.concatenate([corners_collapse, bias_corner], axis=axis)  # (None, shape_, n_dim+1, n_dim+1)

    else:
        bias_corner = o_value + tf.math.reduce_sum(z_value * corners, -2, keepdims=True)
        corners_collapse = K.concatenate([corners_collapse, bias_corner], axis=-2)

    dimensions = np.arange(len(corners.shape))
    if axis != -1:
        dim_permutation = np.concatenate([dimensions[:axis], dimensions[axis + 1 :], [dimensions[axis]]])
    else:
        dim_permutation = np.concatenate([dimensions[:-2], dimensions[-1:], [dimensions[-2]]])

    corners_collapse = K.permute_dimensions(corners_collapse, dim_permutation)
    # tf.linalg.solve works only for float32
    if dtype != dtype32:
        corners_collapse = K.cast(corners_collapse, dtype32)
        corners_pred = K.cast(corners_pred, dtype32)
    w_hull = tf.linalg.solve(matrix=corners_collapse, rhs=K.expand_dims(corners_pred, -1))  # (None, shape_, n_dim+1, 1)

    if dtype != dtype32:
        w_hull = K.cast(w_hull, dtype=dtype)

    shape_prev = np.prod(inputs[-1].shape[1:axis]).astype("int")
    if axis == -1:
        shape_after = 1
    else:
        shape_after = np.prod(inputs[-1].shape[axis + 1 :]).astype("int")
    # (-1, shape_prev, axis, shape_after, n_dim+1)
    flatten_dim = np.prod(w_hull.shape[1:-2])
    w_hull_flat = K.reshape(w_hull, (-1, flatten_dim, n_dim + 1))
    w_u = K.reshape(w_hull_flat[:, :, :-1], (-1, shape_prev, shape_after, n_dim))  # (-1, shape_, n_dim)
    w_u = K.reshape(K.permute_dimensions(w_u, (0, 1, 3, 2)), [-1] + list(inputs[-1].shape[1:]))
    # reshape w_u
    shape_max = K.max(inputs[-1], axis).shape[1:]
    b_u = K.reshape(w_hull_flat[:, :, -1], [-1] + list(shape_max))  # (-1, shape_)

    return [w_u, b_u]


def get_lower_linear_hull_max(
    inputs: List[tf.Tensor],
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    perturbation_domain: Optional[PerturbationDomain] = None,
    axis: int = -1,
    finetune_lower: Optional[tf.Tensor] = None,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Compute the linear hull that overapproximates max along the axis dimension

    Args:
        inputs: list of input tensors
        mode: type of Forward propagation (ibp, affine, or hybrid). Default to hybrid.
        perturbation_domain (optional): type of perturbation domain that encompass the set of perturbations. Defaults to None.
        axis (optional): Defaults to -1. See Keras offical documentation backend.max(., axis)
        finetune_lower: If not None, should be a constant tensor used to fine tune the lower relaxation.

    Raises:
        NotImplementedError: axis <0 and axis!=-1

    Returns:
        list of output tensors. The lower linear relaxation of max(., axis) in the mode format
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)

    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)
    dtype = x.dtype
    input_shape = inputs_outputs_spec.get_input_shape(inputs)

    o_value = K.cast(1.0, dtype)
    z_value = K.cast(0.0, dtype)
    if axis == len(input_shape):
        axis = -1
    if axis != -1 and axis < 0:
        raise NotImplementedError()  # to do

    V = u_c + l_c
    M = -u_c + K.expand_dims(K.max(l_c, axis), axis)
    # M>=0 means that the associated coordinate cannot be an argmax
    # we can use M to add a penalty to the index that cannot be an argmax

    V += K.expand_dims(K.max(K.abs(V), axis), axis) * (K.sign(M) + o_value)

    w_l = K.cast(tf.one_hot(K.argmin(V, axis), depth=V.shape[axis], axis=axis), dtype)
    b_l = K.sum(z_value * w_l, axis)
    # without finetuning: consider a one hot vector with value one at index=argmin(V)

    if finetune_lower is not None:
        alpha = finetune_lower[None]
        y = alpha * u_c + (o_value - alpha) * l_c
        w_l_alpha = K.cast(tf.one_hot(K.argmax(y, axis), depth=V.shape[axis], axis=axis), dtype)
        w_l = (o_value - alpha) * w_l + alpha * w_l_alpha

    return [w_l, b_l]



from tensorflow import Variable
from tensorflow.keras.layers import Layer


def get_weight_index(layer: Layer, weight: Variable) -> int:
    """Get weight index among layer tracked weights

    Args:
        layer: layer we are looking
        weight: weight supposed to be part of tracked weights by the layer

    Returns:
        the index of the weight in `layer.weights` list

    Raises:
        IndexError: if `weight` is not part of `layer.weights`

    """
    indexes = [i for i, w in enumerate(layer.weights) if w is weight]
    try:
        return indexes[0]
    except IndexError:
        raise IndexError(f"The weight {weight} is not tracked by the layer {layer}.")


def get_weight_index_from_name(layer: Layer, weight_name: str) -> int:
    """Get weight index among layer tracked weights

    Args:
        layer: layer we are looking
        weight_name: name of the weight supposed to be part of tracked weights by the layer

    Returns:
        the index of the weight in `layer.weights` list

    Raises:
        AttributeError: if `weight_name` is not the name of an attribute of `layer`
        IndexError: if the corresponding layer attribute is not part of `layer.weights`

    """
    weight = getattr(layer, weight_name)
    try:
        return get_weight_index(layer=layer, weight=weight)
    except IndexError:
        raise IndexError(f"The weight {weight_name} is not tracked by the layer {layer}.")



# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal
from tensorflow.keras.layers import Dense

from decomon.core import ForwardMode, get_affine, get_ibp
from decomon.layers.convert import to_decomon
from decomon.layers.decomon_layers import DecomonDense
from decomon.models.utils import split_activation


def test_DecomonDense_1D_box(n, mode, shared, floatx, decimal, helpers):
    dc_decomp = True
    kwargs_layer = dict(units=1, use_bias=True, dtype=K.floatx())

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # keras layer & function
    keras_layer = Dense(**kwargs_layer)
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)

    # decomon layer & function
    decomon_layer = DecomonDense(dc_decomp=dc_decomp, shared=shared, mode=mode, **kwargs_layer)
    decomon_layer.share_weights(keras_layer)
    outputs = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)

    # test with several kernel values
    kernel_coeffs = [2, -3]
    W_, bias = decomon_layer.get_weights()

    for kernel_coeff in kernel_coeffs:
        # set weights
        if not shared:
            decomon_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])
        keras_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])

        # keras & decomon outputs
        output_ref_ = f_ref(inputs_)
        outputs_ = f_decomon(inputs_)

        # check bounds consistency
        helpers.assert_decomon_layer_output_properties_box(
            full_inputs=inputs_,
            output_ref=output_ref_,
            outputs_for_mode=outputs_,
            dc_decomp=dc_decomp,
            decimal=decimal,
            mode=mode,
        )


def test_DecomonDense_1D_box_nodc(n, helpers):
    dc_decomp = False
    shared = False
    mode = ForwardMode.HYBRID
    decimal = 5
    kwargs_layer = dict(units=1, use_bias=True, dtype=K.floatx())

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # keras layer & function
    keras_layer = Dense(**kwargs_layer)
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)

    #  decomon layer & function
    decomon_layer = DecomonDense(dc_decomp=dc_decomp, shared=shared, mode=mode, **kwargs_layer)
    decomon_layer.share_weights(keras_layer)
    outputs = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)

    #  test with several kernel values
    kernel_coeffs = [2, -3]
    W_, bias = decomon_layer.get_weights()

    for kernel_coeff in kernel_coeffs:
        # set weights
        if not shared:
            decomon_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])
        keras_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])

        # keras & decomon outputs
        output_ref_ = f_ref(inputs_)
        outputs_ = f_decomon(inputs_)

        # check bounds consistency
        helpers.assert_decomon_layer_output_properties_box_linear(
            full_inputs=inputs_,
            output_ref=output_ref_,
            outputs_for_mode=outputs_,
            dc_decomp=dc_decomp,
            decimal=decimal,
            mode=mode,
        )


def test_DecomonDense_multiD_box(odd, mode, dc_decomp, helpers):
    shared = False
    kwargs_layer = dict(units=1, use_bias=True, dtype=K.floatx())
    decimal = 5

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer & function
    keras_layer = Dense(**kwargs_layer)
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)

    # decomon layer & function
    decomon_layer = DecomonDense(dc_decomp=dc_decomp, shared=shared, mode=mode, **kwargs_layer)
    decomon_layer.share_weights(keras_layer)
    outputs = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)

    # test with several kernel values
    kernel_coeffs = [2, -3]
    W_, bias = decomon_layer.get_weights()

    for kernel_coeff in kernel_coeffs:
        # set weights
        if not shared:
            decomon_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])
        keras_layer.set_weights([2 * np.ones_like(W_), np.ones_like(bias)])

        # keras & decomon outputs
        output_ref_ = f_ref(inputs_)
        outputs_ = f_decomon(inputs_)

        # check bounds consistency
        if dc_decomp:
            helpers.assert_decomon_layer_output_properties_box(
                full_inputs=inputs_,
                output_ref=output_ref_,
                outputs_for_mode=outputs_,
                dc_decomp=dc_decomp,
                decimal=decimal,
                mode=mode,
            )
        else:
            helpers.assert_decomon_layer_output_properties_box_linear(
                full_inputs=inputs_,
                output_ref=output_ref_,
                outputs_for_mode=outputs_,
                dc_decomp=dc_decomp,
                decimal=decimal,
                mode=mode,
            )


def test_DecomonDense_1D_to_decomon_box(n, activation, mode, shared, helpers):
    dc_decomp = True
    kwargs_layer = dict(units=1, use_bias=True, dtype=K.floatx())
    decimal = 5
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # keras layer & function & output
    keras_layer = Dense(**kwargs_layer)
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)
    output_ref_ = f_ref(inputs_)

    # decomon layer & function & output via to_decomon
    input_dim = helpers.get_input_dim_from_full_inputs(inputs)
    layers_ref = split_activation(keras_layer)
    decomon_layers = []
    for layer in layers_ref:
        decomon_layers.append(to_decomon(layer, input_dim, dc_decomp=dc_decomp, ibp=ibp, affine=affine, shared=shared))
    outputs = decomon_layers[0](inputs_for_mode)
    if len(decomon_layers) > 1:
        outputs = decomon_layers[1](outputs)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        dc_decomp=dc_decomp,
        decimal=decimal,
        mode=mode,
    )

    # check same weights
    W_, bias = decomon_layers[0].get_weights()
    W_0, b_0 = keras_layer.get_weights()
    assert_almost_equal(W_, W_0, decimal=decimal, err_msg="wrong decomposition")
    assert_almost_equal(bias, b_0, decimal=decimal, err_msg="wrong decomposition")


def test_DecomonDense_multiD_to_decomon_box(odd, activation, mode, dc_decomp, helpers):
    shared = False
    kwargs_layer = dict(units=1, use_bias=True, dtype=K.floatx())
    decimal = 5
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # keras layer & function & output
    keras_layer = Dense(**kwargs_layer)
    output_ref = keras_layer(input_ref)
    W_0, b_0 = keras_layer.get_weights()
    if odd == 0:
        W_0[0] = 1.037377
        W_0[1] = -0.7575816

        keras_layer.set_weights([W_0, b_0])
    if odd == 1:
        W_0[0] = -0.1657672
        W_0[1] = -0.2613032
        W_0[2] = 0.08437371
        keras_layer.set_weights([W_0, b_0])
    f_ref = K.function(inputs, output_ref)
    output_ref_ = f_ref(inputs_)

    #  decomon layer & function via to_decomon
    input_dim = helpers.get_input_dim_from_full_inputs(inputs)
    layers_ref = split_activation(keras_layer)
    decomon_layers = []
    for layer in layers_ref:
        decomon_layers.append(to_decomon(layer, input_dim, dc_decomp=dc_decomp, ibp=ibp, affine=affine, shared=shared))
    outputs = decomon_layers[0](inputs_for_mode)
    if len(decomon_layers) > 1:
        outputs = decomon_layers[1](outputs)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    if dc_decomp:
        helpers.assert_decomon_layer_output_properties_box(
            full_inputs=inputs_,
            output_ref=output_ref_,
            outputs_for_mode=outputs_,
            dc_decomp=dc_decomp,
            decimal=decimal,
            mode=mode,
        )
    else:
        helpers.assert_decomon_layer_output_properties_box_linear(
            full_inputs=inputs_,
            output_ref=output_ref_,
            outputs_for_mode=outputs_,
            dc_decomp=dc_decomp,
            decimal=decimal,
            mode=mode,
        )

    # check same weights
    W_, bias = decomon_layers[0].get_weights()
    W_0, b_0 = keras_layer.get_weights()
    assert_almost_equal(W_, W_0, decimal=decimal, err_msg="wrong decomposition")
    assert_almost_equal(bias, b_0, decimal=decimal, err_msg="wrong decomposition")



import numpy as np
import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Permute, Reshape

from decomon.core import ForwardMode, get_affine, get_ibp
from decomon.layers.convert import to_decomon
from decomon.layers.decomon_reshape import DecomonPermute, DecomonReshape


def keras_target_shape_reshape(input_ref):
    return (np.prod(input_ref.shape[1:]),)


def target_shape_keras2np_reshape(target_shape):
    return (-1, target_shape[0])


def keras_target_shape_permute(input_ref):
    n_dim = len(input_ref.shape) - 1
    return np.random.permutation(n_dim) + 1


def target_shape_keras2np_permute(target_shape):
    return tuple([0] + list(target_shape))


@pytest.mark.parametrize(
    "decomon_layer_class, keras_target_shape_func, target_shape_keras2np_func, np_func",
    [
        (DecomonReshape, keras_target_shape_reshape, target_shape_keras2np_reshape, np.reshape),
        (DecomonPermute, keras_target_shape_permute, target_shape_keras2np_permute, np.transpose),
    ],
)
def test_Decomon_reshape_n_permute_box(
    decomon_layer_class,
    keras_target_shape_func,
    target_shape_keras2np_func,
    np_func,
    mode,
    dc_decomp,
    floatx,
    decimal,
    helpers,
):
    odd, m_0, m_1 = 0, 0, 1
    data_format = "channels_last"

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs=inputs_)

    # target shape
    target_shape = keras_target_shape_func(input_ref)
    target_shape_ = target_shape_keras2np_func(target_shape)

    # original output
    output_ref_ = np_func(input_ref_, target_shape_)

    # decomon layer
    decomon_layer = decomon_layer_class(target_shape, dc_decomp=dc_decomp, mode=mode, dtype=K.floatx())

    # decomon output
    output = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, output)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "keras_target_shape_func, keras_layer_class",
    [
        (keras_target_shape_reshape, Reshape),
        (keras_target_shape_permute, Permute),
    ],
)
def test_Decomon_reshape_n_permute_to_decomon_box(
    keras_target_shape_func, keras_layer_class, shared, floatx, decimal, helpers
):
    odd, m_0, m_1 = 0, 0, 1
    dc_decomp = True
    data_format = "channels_last"
    mode = ForwardMode.HYBRID
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)

    #  target shape
    target_shape = keras_target_shape_func(input_ref)

    # keras layer
    keras_layer = keras_layer_class(target_shape, dtype=K.floatx())

    # original output
    output_ref = keras_layer(input_ref)
    f_ref = K.function(inputs, output_ref)
    output_ref_ = f_ref(inputs_)

    # conversion with to_decomon
    input_dim = helpers.get_input_dim_from_full_inputs(inputs_)
    decomon_layer = to_decomon(keras_layer, input_dim, dc_decomp=dc_decomp, shared=shared, ibp=ibp, affine=affine)

    # decomon outputs
    outputs = decomon_layer(inputs_for_mode)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        dc_decomp=dc_decomp,
        mode=mode,
        decimal=decimal,
    )



import tensorflow as tf
from tensorflow.keras.layers import InputLayer, Lambda


def get_graph_complexity(model: tf.keras.Model) -> int:
    # do not consider Input nodes or Lambda nodes
    # enumerate the number of
    nb_nodes = 0
    depth_keys = list(model._nodes_by_depth.keys())

    for depth in depth_keys:
        nodes = model._nodes_by_depth[depth]
        for node in nodes:
            if isinstance(node.outbound_layer, (InputLayer, Lambda)):
                continue

            nb_nodes += 1

    return nb_nodes



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Input, Layer
from tensorflow.keras.models import Model

from decomon.core import BoxDomain, PerturbationDomain
from decomon.models.models import DecomonModel


class MetricMode(Enum):
    FORWARD = "forward"
    BACKWARD = "backward"


class MetricLayer(ABC, Layer):
    def __init__(
        self,
        ibp: bool,
        affine: bool,
        mode: Union[str, MetricMode],
        perturbation_domain: Optional[PerturbationDomain],
        **kwargs: Any,
    ):
        """
        Args:
            ibp: boolean that indicates whether we propagate constant
                bounds
            affine: boolean that indicates whether we propagate affine
                bounds
            mode: str: 'backward' or 'forward' whether we doforward or
                backward linear relaxation
            perturbation_domain: the type of input perturbation domain for the
                linear relaxation
            **kwargs
        """
        super().__init__(**kwargs)
        self.ibp = ibp
        self.affine = affine
        self.mode = MetricMode(mode)
        self.perturbation_domain: PerturbationDomain
        if perturbation_domain is None:
            self.perturbation_domain = BoxDomain()
        else:
            self.perturbation_domain = perturbation_domain

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "ibp": self.ibp,
                "affine": self.affine,
                "mode": self.mode,
                "perturbation_domain": self.perturbation_domain,
            }
        )
        return config

    @abstractmethod
    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        """
        Args:
            inputs

        Returns:

        """
        pass


class AdversarialCheck(MetricLayer):
    """Training with symbolic LiRPA bounds for promoting adversarial robustness"""

    def __init__(
        self,
        ibp: bool,
        affine: bool,
        mode: Union[str, MetricMode],
        perturbation_domain: Optional[PerturbationDomain],
        **kwargs: Any,
    ):
        """
        Args:
            ibp: boolean that indicates whether we propagate constant
                bounds
            affine: boolean that indicates whether we propagate affine
                bounds
            mode: str: 'backward' or 'forward' whether we doforward or
                backward linear relaxation
            perturbation_domain: the type of input perturbation domain for the
                linear relaxation
            **kwargs
        """
        super().__init__(ibp=ibp, affine=affine, mode=mode, perturbation_domain=perturbation_domain, **kwargs)

    def linear_adv(
        self, z_tensor: tf.Tensor, y_tensor: tf.Tensor, w_u: tf.Tensor, b_u: tf.Tensor, w_l: tf.Tensor, b_l: tf.Tensor
    ) -> tf.Tensor:
        w_upper = w_u * (1 - y_tensor[:, None]) - K.expand_dims(K.sum(w_l * y_tensor[:, None], -1), -1)
        b_upper = b_u * (1 - y_tensor) - b_l * y_tensor

        adv_score = self.perturbation_domain.get_upper(z_tensor, w_upper, b_upper) - 1e6 * y_tensor

        return K.max(adv_score, -1)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        """
        Args:
            inputs

        Returns:
            adv_score <0 if the predictionis robust on the input convex
            domain
        """

        y_tensor = inputs[-1]

        if self.ibp and self.affine:
            _, z, u_c, w_u_f, b_u_f, l_c, w_l_f, b_l_f = inputs[:8]
        elif not self.ibp and self.affine:
            _, z, w_u_f, b_u_f, w_l_f, b_l_f = inputs[:6]
        elif self.ibp and not self.affine:
            _, z, u_c, l_c = inputs[:4]
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if self.ibp:
            adv_ibp = _get_ibp_score(u_c, l_c, y_tensor)
        if self.affine:
            adv_f = _get_affine_score(
                z, w_u_f, b_u_f, w_l_f, b_l_f, y_tensor, perturbation_domain=self.perturbation_domain
            )

        if self.ibp and not self.affine:
            adv_score = adv_ibp
        elif self.ibp and self.affine:
            adv_score = K.minimum(adv_ibp, adv_f)
        elif not self.ibp and self.affine:
            adv_score = adv_f
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if self.mode == MetricMode.BACKWARD:
            w_u_b, b_u_b, w_l_b, b_l_b, _ = inputs[-5:]
            adv_b = _get_backward_score(
                z, w_u_b, b_u_b, w_l_b, b_l_b, y_tensor, perturbation_domain=self.perturbation_domain
            )
            adv_score = K.minimum(adv_score, adv_b)

        return adv_score


class AdversarialScore(AdversarialCheck):
    """Training with symbolic LiRPA bounds for promoting adversarial robustness"""

    def __init__(
        self,
        ibp: bool,
        affine: bool,
        mode: Union[str, MetricMode],
        perturbation_domain: Optional[PerturbationDomain],
        **kwargs: Any,
    ):
        """
        Args:
            ibp: boolean that indicates whether we propagate constant
                bounds
            affine: boolean that indicates whether we propagate affine
                bounds
            mode: str: 'backward' or 'forward' whether we doforward or
                backward linear relaxation
            perturbation_domain: the type of input perturbation domain for the
                linear relaxation
            **kwargs
        """
        super().__init__(ibp=ibp, affine=affine, mode=mode, perturbation_domain=perturbation_domain, **kwargs)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        """
        Args:
            inputs

        Returns:
            adv_score <0 if the predictionis robust on the input convex
            domain
        """

        y_tensor = inputs[-1]

        if self.ibp and self.affine:
            _, z, u_c, w_u_f, b_u_f, l_c, w_l_f, b_l_f = inputs[:8]
        elif not self.ibp and self.affine:
            _, z, w_u_f, b_u_f, w_l_f, b_l_f = inputs[:6]
        elif self.ibp and not self.affine:
            _, z, u_c, l_c = inputs[:4]
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if self.ibp:
            adv_ibp = _get_ibp_score(u_c=l_c, l_c=u_c, source_tensor=y_tensor)
        if self.affine:
            adv_f = _get_affine_score(
                z,
                w_u=w_l_f,
                b_u=b_l_f,
                w_l=w_u_f,
                b_l=b_u_f,
                source_tensor=y_tensor,
                perturbation_domain=self.perturbation_domain,
            )

        if self.ibp and not self.affine:
            adv_score = adv_ibp
        elif self.ibp and self.affine:
            adv_score = K.minimum(adv_ibp, adv_f)
        elif not self.ibp and self.affine:
            adv_score = adv_f
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if self.mode == MetricMode.BACKWARD:
            w_u_b, b_u_b, w_l_b, b_l_b, _ = inputs[-5:]
            adv_b = _get_backward_score(
                z, w_u_b, b_u_b, w_l_b, b_l_b, y_tensor, perturbation_domain=self.perturbation_domain
            )
            adv_score = K.minimum(adv_score, adv_b)

        return adv_score


def build_formal_adv_check_model(decomon_model: DecomonModel) -> tf.keras.Model:
    """automatic design on a Keras  model which predicts a certificate of adversarial robustness

    Args:
        decomon_model

    Returns:

    """
    # check type and that backward pass is available

    perturbation_domain = decomon_model.perturbation_domain
    layer = AdversarialCheck(decomon_model.ibp, decomon_model.affine, decomon_model.mode, perturbation_domain)
    output = decomon_model.output
    input = decomon_model.input
    n_out = decomon_model.output[0].shape[1:]
    y_out = Input(n_out)

    adv_score = layer(output + [y_out])
    adv_model = Model(input + [y_out], adv_score)
    return adv_model


def build_formal_adv_model(decomon_model: DecomonModel) -> tf.keras.Model:
    """automatic design on a Keras  model which predicts a certificate of adversarial robustness

    Args:
        decomon_model

    Returns:

    """
    # check type and that backward pass is available

    perturbation_domain = decomon_model.perturbation_domain
    layer = AdversarialScore(decomon_model.ibp, decomon_model.affine, decomon_model.mode, perturbation_domain)
    output = decomon_model.output
    input = decomon_model.input
    n_out = decomon_model.output[0].shape[1:]
    y_out = Input(n_out)

    adv_score = layer(output + [y_out])
    adv_model = Model(input + [y_out], adv_score)
    return adv_model


class UpperScore(MetricLayer):
    """Training with symbolic LiRPA bounds for limiting the local maximum of a neural network"""

    def __init__(
        self,
        ibp: bool,
        affine: bool,
        mode: Union[str, MetricMode],
        perturbation_domain: Optional[PerturbationDomain],
        **kwargs: Any,
    ):
        """
        Args:
            ibp: boolean that indicates whether we propagate constant
                bounds
            affine: boolean that indicates whether we propagate affine
                bounds
            mode: str: 'backward' or 'forward' whether we doforward or
                backward linear relaxation
            perturbation_domain: the type of input perturbation domain for the
                linear relaxation
            **kwargs
        """
        super().__init__(ibp=ibp, affine=affine, mode=mode, perturbation_domain=perturbation_domain, **kwargs)

    def linear_upper(self, z_tensor: tf.Tensor, y_tensor: tf.Tensor, w_u: tf.Tensor, b_u: tf.Tensor) -> tf.Tensor:
        w_upper = w_u * y_tensor[:, None]
        b_upper = b_u * y_tensor

        upper_score = self.perturbation_domain.get_upper(z_tensor, w_upper, b_upper)

        return K.sum(upper_score, -1)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> tf.Tensor:
        """
        Args:
            inputs

        Returns:
            upper_score <=0 if the maximum of the neural network is
            lower than the target
        """

        y_tensor = inputs[-1]
        z_tensor = inputs[1]

        if self.ibp and self.affine:
            _, _, u_c, w_u_f, b_u_f, _, _, _ = inputs[:8]

            upper_ibp = K.sum(u_c * y_tensor, -1)
            upper_affine = self.linear_upper(z_tensor, y_tensor, w_u_f, b_u_f)
            upper_score = K.minimum(upper_ibp, upper_affine)
        elif not self.ibp and self.affine:
            _, _, w_u_f, b_u_f = inputs[:6]
            upper_score = self.linear_upper(z_tensor, y_tensor, w_u_f, b_u_f)
        elif self.ibp and not self.affine:
            _, _, u_c, l_c = inputs[:4]
            upper_score = K.sum(u_c * y_tensor, -1)
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if self.mode == MetricMode.BACKWARD:
            w_u_b, b_u_b, _, _, _ = inputs[-5:]
            upper_backward = self.linear_upper(z_tensor, y_tensor, w_u_b[:, 0], b_u_b[:, 0])
            upper_score = K.minimum(upper_score, upper_backward)

        return upper_score


def build_formal_upper_model(decomon_model: DecomonModel) -> tf.keras.Model:
    """automatic design on a Keras  model which predicts a certificate on the local upper bound

    Args:
        decomon_model

    Returns:

    """
    # check type and that backward pass is available

    perturbation_domain = decomon_model.perturbation_domain
    layer = UpperScore(decomon_model.ibp, decomon_model.affine, decomon_model.mode, perturbation_domain)
    output = decomon_model.output
    input = decomon_model.input
    n_out = decomon_model.output[0].shape[1:]
    y_out = Input(n_out)

    upper_score = layer(output + [y_out])
    upper_model = Model(input + [y_out], upper_score)
    return upper_model


def _get_ibp_score(
    u_c: tf.Tensor, l_c: tf.Tensor, source_tensor: tf.Tensor, target_tensor: Optional[tf.Tensor] = None
) -> tf.Tensor:
    if target_tensor is None:
        target_tensor = 1.0 - source_tensor

    shape = np.prod(u_c.shape[1:])
    u_c_reshaped = K.reshape(u_c, (-1, shape))
    l_c_reshaped = K.reshape(l_c, (-1, shape))

    score_u = (
        l_c_reshaped * target_tensor
        - K.expand_dims(K.min(u_c_reshaped * source_tensor, -1), -1)
        - 1e6 * (1 - target_tensor)
    )

    return K.max(score_u, -1)


def _get_affine_score(
    z_tensor: tf.Tensor,
    w_u: tf.Tensor,
    b_u: tf.Tensor,
    w_l: tf.Tensor,
    b_l: tf.Tensor,
    source_tensor: tf.Tensor,
    perturbation_domain: PerturbationDomain,
    target_tensor: Optional[tf.Tensor] = None,
) -> tf.Tensor:
    if target_tensor is None:
        target_tensor = 1.0 - source_tensor

    n_dim = w_u.shape[1]
    shape = np.prod(b_u.shape[1:])
    w_u_reshaped = K.reshape(w_u, (-1, n_dim, shape, 1))
    w_l_reshaped = K.reshape(w_l, (-1, n_dim, 1, shape))
    b_u_reshaped = K.reshape(b_u, (-1, shape, 1))
    b_l_reshaped = K.reshape(b_l, (-1, 1, shape))

    w_u_f = w_l_reshaped - w_u_reshaped
    b_u_f = b_l_reshaped - b_u_reshaped

    # add penalties on biases
    b_u_f = b_u_f - 1e6 * (1 - target_tensor)[:, None, :]
    b_u_f = b_u_f - 1e6 * (1 - source_tensor)[:, :, None]

    upper = perturbation_domain.get_upper(z_tensor, w_u_f, b_u_f)
    return K.max(upper, (-1, -2))


def _get_backward_score(
    z_tensor: tf.Tensor,
    w_u: tf.Tensor,
    b_u: tf.Tensor,
    w_l: tf.Tensor,
    b_l: tf.Tensor,
    source_tensor: tf.Tensor,
    perturbation_domain: PerturbationDomain,
    target_tensor: Optional[tf.Tensor] = None,
) -> tf.Tensor:
    return _get_affine_score(
        z_tensor,
        w_u[:, 0],
        b_u[:, 0],
        w_l[:, 0],
        b_l[:, 0],
        source_tensor,
        perturbation_domain=perturbation_domain,
        target_tensor=target_tensor,
    )



import tensorflow.keras.backend as K

from decomon.metrics.utils import categorical_cross_entropy


def test_categorical_cross_entropy(odd, mode, floatx, decimal, helpers):
    dc_decomp = False

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_multid_box(odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_multid_box(odd, dc_decomp=dc_decomp)

    # original output
    f_ref = K.function(inputs, -input_ref + K.log(K.sum(K.exp(input_ref), -1))[:, None])
    output_ref_ = f_ref(inputs_)

    # decomon output
    output = categorical_cross_entropy(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    f_entropy = K.function(inputs, output)
    outputs_ = f_entropy(inputs_)

    # check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        dc_decomp=dc_decomp,
        mode=mode,
        decimal=decimal,
    )



from typing import Any, Dict, List, Optional, Tuple, Union

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import (
    Add,
    Average,
    Concatenate,
    Dot,
    Lambda,
    Maximum,
    Minimum,
    Multiply,
    Subtract,
)

from decomon.core import ForwardMode, PerturbationDomain
from decomon.layers.core import DecomonLayer
from decomon.layers.utils import broadcast, multiply, permute_dimensions
from decomon.utils import maximum, minus, subtract

##### Merge Layer ####


class DecomonMerge(DecomonLayer):
    """Base class for Decomon layers based on Mergind Keras layers."""

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> List[tf.TensorShape]:
        return input_shape

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        n_comp = self.nb_tensors
        input_shape_y = input_shape[n_comp - 1 :: n_comp]
        self.original_keras_layer_class.build(self, input_shape_y)


class DecomonAdd(DecomonMerge, Add):
    """LiRPA implementation of Add layers.
    See Keras official documentation for further details on the Add operator

    """

    original_keras_layer_class = Add

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        # splits the inputs
        (
            inputs_x,
            inputs_u_c,
            inputs_w_u,
            inputs_b_u,
            inputs_l_c,
            inputs_w_l,
            inputs_b_l,
            inputs_h,
            inputs_g,
        ) = self.inputs_outputs_spec.get_fullinputs_by_type_from_inputsformode_to_merge(inputs)
        x_out = inputs_x[0]
        dtype = x_out.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        # outputs
        if self.ibp:
            u_c_out = sum(inputs_u_c)
            l_c_out = sum(inputs_l_c)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = sum(inputs_b_u)
            b_l_out = sum(inputs_b_l)
            w_u_out = sum(inputs_w_u)
            w_l_out = sum(inputs_w_l)
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_out, g_out = empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x_out, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )


class DecomonAverage(DecomonMerge, Average):
    """LiRPA implementation of Average layers.
    See Keras official documentation for further details on the Average operator

    """

    original_keras_layer_class = Average

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )
        self.op = Lambda(lambda x: sum(x) / len(x))

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        # splits the inputs
        (
            inputs_x,
            inputs_u_c,
            inputs_w_u,
            inputs_b_u,
            inputs_l_c,
            inputs_w_l,
            inputs_b_l,
            inputs_h,
            inputs_g,
        ) = self.inputs_outputs_spec.get_fullinputs_by_type_from_inputsformode_to_merge(inputs)
        x_out = inputs_x[0]
        dtype = x_out.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        # outputs
        if self.ibp:
            u_c_out = self.op(inputs_u_c)
            l_c_out = self.op(inputs_l_c)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = self.op(inputs_b_u)
            b_l_out = self.op(inputs_b_l)
            w_u_out = self.op(inputs_w_u)
            w_l_out = self.op(inputs_w_l)
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_out, g_out = empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x_out, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )


class DecomonSubtract(DecomonMerge, Subtract):
    """LiRPA implementation of Subtract layers.
    See Keras official documentation for further details on the Subtract operator

    """

    original_keras_layer_class = Subtract

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.dc_decomp:
            raise NotImplementedError()

        # splits the inputs
        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)

        # check number of inputs
        if len(inputs_list) != 2:
            raise ValueError("This layer is intended to merge only 2 layers.")

        output = subtract(
            inputs_list[0],
            inputs_list[1],
            dc_decomp=self.dc_decomp,
            perturbation_domain=self.perturbation_domain,
            mode=self.mode,
        )
        return output


class DecomonMinimum(DecomonMerge, Minimum):
    """LiRPA implementation of Minimum layers.
    See Keras official documentation for further details on the Minimum operator

    """

    original_keras_layer_class = Minimum

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.dc_decomp:
            raise NotImplementedError()

        # splits the inputs
        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)
        # look at minus the input to apply maximum
        inputs_list = [
            minus(single_inputs, mode=self.mode, dc_decomp=self.dc_decomp, perturbation_domain=self.perturbation_domain)
            for single_inputs in inputs_list
        ]

        #  check number of inputs
        if len(inputs_list) == 1:  # nothing to merge
            return inputs
        else:
            output = maximum(
                inputs_list[0],
                inputs_list[1],
                dc_decomp=self.dc_decomp,
                perturbation_domain=self.perturbation_domain,
                mode=self.mode,
            )
            for j in range(2, len(inputs_list)):
                output = maximum(
                    output,
                    inputs_list[j],
                    dc_decomp=self.dc_decomp,
                    perturbation_domain=self.perturbation_domain,
                    mode=self.mode,
                )

            return minus(output, mode=self.mode, dc_decomp=self.dc_decomp, perturbation_domain=self.perturbation_domain)


class DecomonMaximum(DecomonMerge, Maximum):
    """LiRPA implementation of Maximum layers.
    See Keras official documentation for further details on the Maximum operator

    """

    original_keras_layer_class = Maximum

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.dc_decomp:
            raise NotImplementedError()

        # splits the inputs
        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)

        #  check number of inputs
        if len(inputs_list) == 1:  # nothing to merge
            return inputs
        else:
            output = maximum(
                inputs_list[0],
                inputs_list[1],
                dc_decomp=self.dc_decomp,
                perturbation_domain=self.perturbation_domain,
                mode=self.mode,
            )
            for j in range(2, len(inputs_list)):
                output = maximum(
                    output,
                    inputs_list[j],
                    dc_decomp=self.dc_decomp,
                    perturbation_domain=self.perturbation_domain,
                    mode=self.mode,
                )

        return output


class DecomonConcatenate(DecomonMerge, Concatenate):
    """LiRPA implementation of Concatenate layers.
    See Keras official documentation for further details on the Concatenate operator

    """

    original_keras_layer_class = Concatenate

    def __init__(
        self,
        axis: int = -1,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            axis=axis,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

        def func(inputs: List[tf.Tensor]) -> tf.Tensor:
            return Concatenate.call(self, inputs)

        self.op = func
        if self.axis == -1:
            self.op_w = self.op
        else:
            self.op_w = Concatenate(axis=self.axis + 1)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        # splits the inputs
        (
            inputs_x,
            inputs_u_c,
            inputs_w_u,
            inputs_b_u,
            inputs_l_c,
            inputs_w_l,
            inputs_b_l,
            inputs_h,
            inputs_g,
        ) = self.inputs_outputs_spec.get_fullinputs_by_type_from_inputsformode_to_merge(inputs)
        x_out = inputs_x[0]
        dtype = x_out.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        # outputs
        if self.ibp:
            u_c_out = self.op(inputs_u_c)
            l_c_out = self.op(inputs_l_c)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            b_u_out = self.op(inputs_b_u)
            b_l_out = self.op(inputs_b_l)
            w_u_out = self.op_w(inputs_w_u)
            w_l_out = self.op_w(inputs_w_l)
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_out, g_out = empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x_out, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )


class DecomonMultiply(DecomonMerge, Multiply):
    """LiRPA implementation of Multiply layers.
    See Keras official documentation for further details on the Multiply operator

    """

    original_keras_layer_class = Multiply

    def __init__(
        self,
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.dc_decomp:
            raise NotImplementedError()

        # splits the inputs
        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)

        #  check number of inputs
        if len(inputs_list) == 1:  # nothing to merge
            return inputs
        else:
            output = multiply(
                inputs_list[0],
                inputs_list[1],
                dc_decomp=self.dc_decomp,
                perturbation_domain=self.perturbation_domain,
                mode=self.mode,
            )
            for j in range(2, len(inputs_list)):
                output = multiply(
                    output,
                    inputs_list[j],
                    dc_decomp=self.dc_decomp,
                    perturbation_domain=self.perturbation_domain,
                    mode=self.mode,
                )

            return output


class DecomonDot(DecomonMerge, Dot):
    """LiRPA implementation of Dot layers.
    See Keras official documentation for further details on the Dot operator

    """

    original_keras_layer_class = Dot

    def __init__(
        self,
        axes: Union[int, Tuple[int, int]] = (-1, -1),
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        finetune: bool = False,
        shared: bool = False,
        fast: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            axes=axes,
            perturbation_domain=perturbation_domain,
            dc_decomp=dc_decomp,
            mode=mode,
            finetune=finetune,
            shared=shared,
            fast=fast,
            **kwargs,
        )
        if isinstance(axes, int):
            self.axes = (axes, axes)
        else:
            self.axes = axes

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        if self.dc_decomp:
            raise NotImplementedError()

        # splits the inputs
        inputs_list = self.inputs_outputs_spec.split_inputsformode_to_merge(inputs)

        #  check number of inputs
        if len(inputs_list) == 1:  # nothing to merge
            return inputs
        elif len(inputs_list) == 2:
            inputs_0, inputs_1 = inputs_list
        else:
            raise NotImplementedError("This layer is not implemented to merge more than 2 layers.")

        input_shape_0 = self.inputs_outputs_spec.get_input_shape(inputs_0)
        input_shape_1 = self.inputs_outputs_spec.get_input_shape(inputs_1)
        n_0 = len(input_shape_0) - 2
        n_1 = len(input_shape_1) - 2

        inputs_0 = permute_dimensions(inputs_0, self.axes[0], mode=self.mode, dc_decomp=self.dc_decomp)
        inputs_1 = permute_dimensions(inputs_1, self.axes[1], mode=self.mode, dc_decomp=self.dc_decomp)
        inputs_0 = broadcast(
            inputs_0, n_1, -1, mode=self.mode, dc_decomp=self.dc_decomp, perturbation_domain=self.perturbation_domain
        )
        inputs_1 = broadcast(
            inputs_1, n_0, 2, mode=self.mode, dc_decomp=self.dc_decomp, perturbation_domain=self.perturbation_domain
        )
        outputs_multiply = multiply(
            inputs_0, inputs_1, dc_decomp=self.dc_decomp, perturbation_domain=self.perturbation_domain, mode=self.mode
        )

        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = self.inputs_outputs_spec.get_fullinputs_from_inputsformode(
            outputs_multiply, compute_ibp_from_affine=False
        )
        dtype = x.dtype
        empty_tensor = self.inputs_outputs_spec.get_empty_tensor(dtype=dtype)

        if self.ibp:
            u_c_out = K.sum(u_c, 1)
            l_c_out = K.sum(l_c, 1)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.affine:
            w_u_out = K.sum(w_u, 2)
            b_u_out = K.sum(b_u, 1)
            w_l_out = K.sum(w_l, 2)
            b_l_out = K.sum(b_l, 1)
        else:
            w_u_out, b_u_out, w_l_out, b_l_out = empty_tensor, empty_tensor, empty_tensor, empty_tensor

        if self.dc_decomp:
            raise NotImplementedError()
        else:
            h_out, g_out = empty_tensor, empty_tensor

        return self.inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out, h_out, g_out]
        )



import pytest

from decomon.core import ForwardMode
from decomon.layers.decomon_layers import (
    DecomonActivation,
    DecomonBatchNormalization,
    DecomonConv2D,
    DecomonDense,
    DecomonDropout,
    DecomonFlatten,
    DecomonInputLayer,
)
from decomon.layers.decomon_merge_layers import (
    DecomonAdd,
    DecomonAverage,
    DecomonConcatenate,
    DecomonDot,
    DecomonMaximum,
    DecomonMinimum,
    DecomonMultiply,
    DecomonSubtract,
)
from decomon.layers.decomon_reshape import DecomonPermute, DecomonReshape
from decomon.layers.maxpooling import DecomonMaxPooling2D

try:
    from decomon.layers.deel_lip import DecomonGroupSort, DecomonGroupSort2
except ImportError:
    deel_lip_available = False
else:
    deel_lip_available = True

deel_lip_skip_reason = "deel-lip is not available"


def test_decomon_reshape():
    dims = (1, 2, 3)
    mode = ForwardMode.AFFINE
    layer = DecomonPermute(dims=dims, mode=mode)
    config = layer.get_config()
    assert config["dims"] == dims
    assert config["mode"] == mode

    shape = (1, 2, 3)
    layer = DecomonReshape(target_shape=shape, mode=mode)
    config = layer.get_config()
    assert config["target_shape"] == shape
    assert config["mode"] == mode


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_deel_lip():
    layer = DecomonGroupSort()
    config = layer.get_config()

    layer = DecomonGroupSort2()
    config = layer.get_config()


def test_maxpooling():
    layer = DecomonMaxPooling2D()
    config = layer.get_config()
    assert "pool_size" in config


def test_decomon_merge_layers():
    for cls in [
        DecomonAdd,
        DecomonAverage,
        DecomonConcatenate,
        DecomonDot,
        DecomonMaximum,
        DecomonMinimum,
        DecomonMultiply,
        DecomonSubtract,
    ]:
        layer = cls()
        config = layer.get_config()
        assert "mode" in config
        if layer == DecomonConcatenate:
            assert "axis" in config
        elif layer == DecomonDot:
            assert "axes" in config


def test_decomon_layers():
    activation_mapping = {
        "relu": "relu",
        "softmax": "softmax",
        "linear": "linear",
        None: "linear",
    }
    for activation_arg, activation_res in activation_mapping.items():
        layer = DecomonActivation(activation=activation_arg)
        config = layer.get_config()
        print(config)
        print(layer.activation)
        assert config["activation"] == activation_res

    units = 2
    layer = DecomonDense(units=units)
    config = layer.get_config()
    assert config["units"] == units

    filters = 2
    kernel_size = 3, 3
    layer = DecomonConv2D(filters=filters, kernel_size=kernel_size)
    config = layer.get_config()
    assert config["filters"] == filters
    assert config["kernel_size"] == kernel_size

    rate = 0.9
    layer = DecomonDropout(rate=rate)
    config = layer.get_config()
    assert config["rate"] == rate

    for cls in [DecomonBatchNormalization, DecomonInputLayer, DecomonFlatten]:
        layer = cls()
        config = layer.get_config()
        assert "mode" in config
        if layer == DecomonBatchNormalization:
            assert "axis" in config
        elif layer == DecomonInputLayer:
            assert "input_shape" in config
        elif layer == DecomonFlatten:
            assert "data_format" in config



import numpy as np
import pytest
import tensorflow as tf
from numpy.testing import assert_almost_equal
from tensorflow.keras import Input
from tensorflow.keras.layers import Activation, Conv2D, Dense, Layer, Permute, PReLU

from decomon.models.utils import (
    convert_deellip_to_keras,
    preprocess_layer,
    split_activation,
)

try:
    import deel.lip
except ImportError:
    deel_lip_available = False
else:
    deel_lip_available = True
    from deel.lip.activations import GroupSort
    from deel.lip.layers import SpectralDense

deel_lip_skip_reason = "deel-lip is not available"


@pytest.mark.parametrize(
    "layer_class, layer_kwargs, input_shape_wo_batchsize, embedded_activation_layer_class",
    [
        (Dense, {"units": 3, "activation": "relu"}, (1,), None),
        (Dense, {"units": 3}, (1,), PReLU),
        (Conv2D, {"filters": 2, "kernel_size": (3, 3), "activation": "relu"}, (64, 64, 3), None),
    ],
)
def test_split_activation_do_split(
    layer_class, layer_kwargs, input_shape_wo_batchsize, embedded_activation_layer_class, use_bias
):
    # init layer
    if embedded_activation_layer_class is not None:
        layer = layer_class(use_bias=use_bias, activation=embedded_activation_layer_class(), **layer_kwargs)
    else:
        layer = layer_class(use_bias=use_bias, **layer_kwargs)
    # init input_shape and weights
    input_shape = input_shape_wo_batchsize
    input_tensor = Input(input_shape)
    layer(input_tensor)
    # split
    layers = split_activation(layer)
    # check layer split
    assert len(layers) == 2
    layer_wo_activation, activation_layer = layers
    assert isinstance(layer_wo_activation, layer.__class__)
    assert layer_wo_activation.get_config()["activation"] == "linear"
    if isinstance(layer.activation, Layer):
        assert activation_layer == layer.activation
    else:
        assert isinstance(activation_layer, Activation)
        assert activation_layer.get_config()["activation"] == layer_kwargs["activation"]
    # check names starts with with original name + "_"
    assert layer_wo_activation.name.startswith(f"{layer.name}_")
    assert activation_layer.name.startswith(f"{layer.name}_")
    # check already built
    assert layer_wo_activation.built
    assert activation_layer.built
    # check same outputs
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = layer(inputs_np).numpy()
    output_np_new = activation_layer(layer_wo_activation(inputs_np)).numpy()
    assert_almost_equal(output_np_new, output_np_ref)
    # check same trainable weights
    for i in range(len(layer._trainable_weights)):
        assert layer._trainable_weights[i] is layer_wo_activation._trainable_weights[i]


@pytest.mark.parametrize(
    "layer_class, layer_kwargs",
    [
        (Dense, {"units": 3}),
        (Activation, {"activation": "relu"}),
        (Permute, {"dims": (1, 2, 3)}),
    ],
)
def test_split_activation_do_nothing(layer_class, layer_kwargs):
    layer = layer_class(**layer_kwargs)
    layers = split_activation(layer)
    assert len(layers) == 1
    assert layers[0] == layer


def test_split_activation_uninitialized_layer_ko():
    layer = Dense(3, activation="relu")
    with pytest.raises(RuntimeError):
        layers = split_activation(layer)


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
@pytest.mark.parametrize(
    "layer_class_name, "
    "layer_kwargs, "
    "input_shape_wo_batchsize, "
    "embedded_activation_layer_class_name, "
    "embedded_activation_layer_kwargs",
    [
        ("SpectralDense", {"units": 3, "activation": "relu"}, (1,), None, None),
        ("Dense", {"units": 3}, (1,), "GroupSort", {"n": 1}),
    ],
)
def test_split_activation_do_split_with_deellip(
    layer_class_name,
    layer_kwargs,
    input_shape_wo_batchsize,
    embedded_activation_layer_class_name,
    embedded_activation_layer_kwargs,
    use_bias,
):
    # init layer
    layer_class = globals()[layer_class_name]
    if embedded_activation_layer_class_name is not None:
        embedded_activation_layer_class = globals()[embedded_activation_layer_class_name]
        layer = layer_class(
            use_bias=use_bias,
            activation=embedded_activation_layer_class(**embedded_activation_layer_kwargs),
            **layer_kwargs,
        )
    else:
        layer = layer_class(use_bias=use_bias, **layer_kwargs)
    # init input_shape and weights
    input_shape = input_shape_wo_batchsize
    input_tensor = Input(input_shape)
    layer(input_tensor)
    # split
    layers = split_activation(layer)
    # check layer split
    assert len(layers) == 2
    layer_wo_activation, activation_layer = layers
    assert isinstance(layer_wo_activation, layer.__class__)
    assert layer_wo_activation.get_config()["activation"] == "linear"
    if isinstance(layer.activation, Layer):
        assert activation_layer == layer.activation
    else:
        assert isinstance(activation_layer, Activation)
        assert activation_layer.get_config()["activation"] == layer_kwargs["activation"]
    # check names starts with with original name + "_"
    assert layer_wo_activation.name.startswith(f"{layer.name}_")
    assert activation_layer.name.startswith(f"{layer.name}_")
    # check already built
    assert layer_wo_activation.built
    assert activation_layer.built
    # check same outputs
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = layer(inputs_np).numpy()
    output_np_new = activation_layer(layer_wo_activation(inputs_np)).numpy()
    assert_almost_equal(output_np_new, output_np_ref)
    # check same trainable weights
    for i in range(len(layer._trainable_weights)):
        assert layer._trainable_weights[i] is layer_wo_activation._trainable_weights[i]


def test_convert_deellip_to_keras_dense():
    layer = Dense(units=3)
    keras_layer = convert_deellip_to_keras(layer)
    # new attributes added
    assert hasattr(keras_layer, "is_lipschitz")
    assert hasattr(keras_layer, "deellip_classname")
    assert hasattr(keras_layer, "k_coef_lip")
    # check values
    assert not keras_layer.is_lipschitz
    assert keras_layer.deellip_classname == layer.__class__.__name__
    assert keras_layer.k_coef_lip == -1.0
    assert keras_layer is layer
    assert keras_layer.name.startswith(layer.name)
    # idempotency
    keras_layer2 = convert_deellip_to_keras(keras_layer)
    assert keras_layer is keras_layer2
    assert keras_layer.is_lipschitz == keras_layer2.is_lipschitz
    assert keras_layer.deellip_classname == keras_layer2.deellip_classname
    assert keras_layer.k_coef_lip == keras_layer2.k_coef_lip


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_convert_deellip_to_keras_groupsort():
    layer = GroupSort(n=2)
    keras_layer = convert_deellip_to_keras(layer)
    # new attributes added
    assert hasattr(keras_layer, "is_lipschitz")
    assert hasattr(keras_layer, "deellip_classname")
    assert hasattr(keras_layer, "k_coef_lip")
    # check values
    assert keras_layer.is_lipschitz
    assert keras_layer.deellip_classname == layer.__class__.__name__
    assert keras_layer.k_coef_lip == layer.k_coef_lip
    assert keras_layer is layer
    assert keras_layer.name.startswith(layer.name)
    # idempotency
    keras_layer2 = convert_deellip_to_keras(keras_layer)
    assert keras_layer is keras_layer2
    assert keras_layer.is_lipschitz == keras_layer2.is_lipschitz
    assert keras_layer.deellip_classname == keras_layer2.deellip_classname
    assert keras_layer.k_coef_lip == keras_layer2.k_coef_lip


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_convert_deellip_to_keras_spectraldense():
    units = 3
    layer = SpectralDense(units=units)
    layer(
        Input(
            1,
        )
    )
    keras_layer = convert_deellip_to_keras(layer)
    # new attributes added
    assert hasattr(keras_layer, "is_lipschitz")
    assert hasattr(keras_layer, "deellip_classname")
    assert hasattr(keras_layer, "k_coef_lip")
    # check values
    assert keras_layer.is_lipschitz
    assert keras_layer.deellip_classname == layer.__class__.__name__
    assert keras_layer.k_coef_lip == layer.k_coef_lip
    assert isinstance(keras_layer, Dense)
    assert keras_layer.units == units
    assert keras_layer.name.startswith(layer.name)
    # same output?
    input_tensor = tf.ones((4, 1))
    output_ref = layer(input_tensor).numpy()
    new_output = keras_layer(input_tensor).numpy()
    assert_almost_equal(new_output, output_ref)
    # idempotency
    keras_layer2 = convert_deellip_to_keras(keras_layer)
    assert keras_layer is keras_layer2
    assert keras_layer.is_lipschitz == keras_layer2.is_lipschitz
    assert keras_layer.deellip_classname == keras_layer2.deellip_classname
    assert keras_layer.k_coef_lip == keras_layer2.k_coef_lip


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_convert_deellip_to_keras_spectraldense_not_initialized_ko():
    units = 3
    layer = SpectralDense(units=units)
    with pytest.raises(RuntimeError):
        keras_layer = convert_deellip_to_keras(layer)


@pytest.mark.parametrize(
    "layer_class_name, layer_kwargs, input_shape_wo_batchsize, is_deel_lip",
    [
        ("Dense", {"units": 3}, (1,), False),
        ("Activation", {"activation": "relu"}, (1,), False),
        ("Permute", {"dims": (1, 2, 3)}, (1, 1, 1), False),
        ("SpectralDense", {"units": 3}, (1,), True),
        ("GroupSort", {"n": 1}, (1,), True),
    ],
)
def test_preprocess_layer_no_nonlinear_activation(
    layer_class_name, layer_kwargs, input_shape_wo_batchsize, is_deel_lip
):
    if is_deel_lip and not deel_lip_available:
        pytest.skip(deel_lip_skip_reason)
    layer_class = globals()[layer_class_name]
    layer = layer_class(**layer_kwargs)
    # build layer
    input_tensor = Input(input_shape_wo_batchsize)
    layer(input_tensor)
    # preprocess
    layers = preprocess_layer(layer)
    # check resulting layers
    assert len(layers) == 1
    keras_layer = layers[0]
    # new attributes added
    assert hasattr(keras_layer, "is_lipschitz")
    assert hasattr(keras_layer, "deellip_classname")
    assert hasattr(keras_layer, "k_coef_lip")
    # check values
    assert keras_layer.deellip_classname == layer.__class__.__name__
    if is_deel_lip:
        assert keras_layer.is_lipschitz
        assert keras_layer.k_coef_lip == layer.k_coef_lip
        if hasattr(layer, "vanilla_export"):
            assert keras_layer is not layer
        else:
            assert keras_layer is layer
    else:
        assert not keras_layer.is_lipschitz
        assert keras_layer.k_coef_lip == -1.0
        assert keras_layer is layer
    # try to call the resulting layers 3 times
    input_tensor = tf.ones((5,) + input_shape_wo_batchsize)
    for _ in range(3):
        keras_layer(input_tensor)


@pytest.mark.parametrize(
    "layer_class_name, "
    "layer_kwargs, "
    "input_shape_wo_batchsize, "
    "embedded_activation_layer_class_name, "
    "embedded_activation_layer_class_kwargs, "
    "is_deellip_layer, "
    "is_deellip_activation",
    [
        ("Dense", {"units": 3, "activation": "relu"}, (1,), None, None, False, False),
        ("Dense", {"units": 3}, (1,), "PReLU", {}, False, False),
        ("Conv2D", {"filters": 2, "kernel_size": (3, 3), "activation": "relu"}, (64, 64, 3), None, None, False, False),
        ("Dense", {"units": 3}, (1,), "GroupSort", {"n": 1}, False, True),
        ("SpectralDense", {"units": 3, "activation": "relu"}, (1,), None, None, True, False),
        ("SpectralDense", {"units": 3}, (1,), "PReLU", {}, True, False),
        ("SpectralDense", {"units": 3}, (1,), "GroupSort", {"n": 1}, True, True),
    ],
)
def test_preprocess_layer_nonlinear_activation(
    layer_class_name,
    layer_kwargs,
    input_shape_wo_batchsize,
    embedded_activation_layer_class_name,
    embedded_activation_layer_class_kwargs,
    use_bias,
    is_deellip_layer,
    is_deellip_activation,
):
    # skip deel-lip cases if not available
    if not deel_lip_available and (is_deellip_layer or is_deellip_activation):
        pytest.skip(reason=deel_lip_skip_reason)
    # init layer
    layer_class = globals()[layer_class_name]
    if embedded_activation_layer_class_name is not None:
        embedded_activation_layer_class = globals()[embedded_activation_layer_class_name]
        embedded_activation_layer = embedded_activation_layer_class(**embedded_activation_layer_class_kwargs)
        layer = layer_class(use_bias=use_bias, activation=embedded_activation_layer, **layer_kwargs)
    else:
        layer = layer_class(use_bias=use_bias, **layer_kwargs)
    # init input_shape and weights
    input_shape = input_shape_wo_batchsize
    input_tensor = Input(input_shape)
    layer(input_tensor)
    # split
    layers = preprocess_layer(layer)
    # check layer split
    assert len(layers) == 2
    layer_wo_activation, activation_layer = layers
    if is_deellip_layer:
        if layer.__class__.__name__.endswith("Dense"):
            assert isinstance(layer_wo_activation, Dense)
        if layer.__class__.__name__.endswith("Conv2D"):
            assert isinstance(layer_wo_activation, Conv2D)
    else:
        assert isinstance(layer_wo_activation, layer.__class__)
    assert layer_wo_activation.get_config()["activation"] == "linear"
    if isinstance(layer.activation, Layer):
        assert activation_layer == layer.activation
    else:
        assert isinstance(activation_layer, Activation)
        assert activation_layer.get_config()["activation"] == layer_kwargs["activation"]
    # check names starts with with original name + "_"
    assert layer_wo_activation.name.startswith(f"{layer.name}_")
    assert activation_layer.name.startswith(f"{layer.name}_")
    # check already built
    assert layer_wo_activation.built
    assert activation_layer.built
    # check same outputs
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = layer(inputs_np).numpy()
    output_np_new = activation_layer(layer_wo_activation(inputs_np)).numpy()
    assert_almost_equal(output_np_new, output_np_ref)
    # check same trainable weights
    if not is_deellip_layer:
        for i in range(len(layer._trainable_weights)):
            assert layer._trainable_weights[i] is layer_wo_activation._trainable_weights[i]
    # check deel-lip attributes
    for i, keras_layer in enumerate(layers):
        # new attributes added
        assert hasattr(keras_layer, "is_lipschitz")
        assert hasattr(keras_layer, "deellip_classname")
        assert hasattr(keras_layer, "k_coef_lip")
        # check values
        if i == 0:
            assert keras_layer.deellip_classname == layer.__class__.__name__
            if is_deellip_layer:
                assert keras_layer.is_lipschitz
                assert keras_layer.k_coef_lip == layer.k_coef_lip
            else:
                assert not keras_layer.is_lipschitz
                assert keras_layer.k_coef_lip == -1.0
        else:
            if isinstance(layer.activation, Layer):
                assert keras_layer.deellip_classname == layer.activation.__class__.__name__
            else:
                assert keras_layer.deellip_classname == Activation.__name__
            if is_deellip_activation:
                assert keras_layer.is_lipschitz
                assert keras_layer.k_coef_lip == layer.activation.k_coef_lip
            else:
                assert not keras_layer.is_lipschitz
                assert keras_layer.k_coef_lip == -1.0
